"""
Parsing of FAST output files, binary or text.
Binary read based on Matlab function ReadFASTbinary.m by Bonnie Jonkman.
Author: Evan Gaertner, National Renewable Energy Laboratory

Inputs:
    FileName      - string: contains file name to open
    OutFileFmt    - int: (optional) 1=textfile, 2=binary

Outputs:
    data          - dict: FAST output time series, output channel names as dict keys
    meta          - dict: additional meta data from output file, keys include: 'units', 'DescStr', 'FileID'

"""
import numpy as np
import struct
import os

def ReadFASToutFormat(FileName, OutFileFmt=0, Verbose=False):
    
    if OutFileFmt == 2:
        path,fname = os.path.split(FileName)
        FileName = os.path.join(path, '.'.join(fname.split('.')[:-1])+'.outb')
        Channels, ChanName, ChanUnit, FileID, DescStr = ReadFASTbinary(FileName)
    elif OutFileFmt == 1: 
        path,fname = os.path.split(FileName)
        FileName = os.path.join(path, '.'.join(fname.split('.')[:-1])+'.out')
        Channels, ChanName, ChanUnit, FileID, DescStr = ReadFASTtext(FileName)
    else:
        if Verbose:
            print('Attempting to read FAST output file: %s, format not specified'%FileName)
        error = False
        try:
            if Verbose:
                print('Attempting binary read')
            path,fname = os.path.split(FileName)
            FileName = os.path.join(path, '.'.join(fname.split('.')[:-1])+'.outb')
            Channels, ChanName, ChanUnit, FileID, DescStr = ReadFASTbinary(FileName)
            if Verbose:
                print('Success')
            error = False
        except:
            if Verbose:
                print('Failed')
            error = True
        if error:
            try:
                if Verbose:
                    print('Attempting text read')
                path,fname = os.path.split(FileName)
                FileName = os.path.join(path, '.'.join(fname.split('.')[:-1])+'.out')
                Channels, ChanName, ChanUnit, FileID, DescStr = ReadFASTtext(FileName)
                if Verbose:
                    print('Success')
                error = False
            except:
                if Verbose:
                    print('Failed')
                error = True
        if error:
            raise NameError('Unable read FAST output file: %s'%FileName)

    data = {}
    meta = {}
    meta['units'] = {}
    for i, (chan, unit) in enumerate(zip(ChanName, ChanUnit)):
        data[chan] = Channels[:,i]
        meta['units'][chan] = unit
    meta['FileID'] = FileID
    meta['DescStr'] = DescStr

    return data, meta

def ReadFASTbinary(FileName):
    LenName = 10    # number of characters per channel name
    LenUnit = 10    # number of characters per unit name

    #----------------------------        
    # load file binary data
    #----------------------------
    f = open(FileName, 'rb')
    data = f.read()
    f.close()
    i = 0           # data position pointer for binary unpacking

    #----------------------------        
    # get the header information
    #----------------------------

    FileID = struct.unpack('h',data[i:i+2])[0]              # FAST output file format, INT(2)
    i+=2
    NumOutChans = struct.unpack('i',data[i:i+4])[0]         # The number of output channels, INT(4)
    i+=4
    NT = struct.unpack('i',data[i:i+4])[0]                  # The number of time steps, INT(4)
    i+=4

    if FileID == 1: # with time
        TimeScl = struct.unpack('d',data[i:i+8])[0]         # The time slopes for scaling, REAL(8)
        i+=8
        TimeOff = struct.unpack('d',data[i:i+8])[0]         # The time offsets for scaling, REAL(8)
        i+=8
    else: # without time
        TimeOut1 = struct.unpack('d',data[i:i+8])[0]        # The first time in the time series, REAL(8)
        i+=8
        TimeIncr = struct.unpack('d',data[i:i+8])[0]        # The time increment, REAL(8)
        i+=8
    
    ColScl = [None]*NumOutChans
    for idx in range(0,NumOutChans):
        ColScl[idx] = struct.unpack('f',data[i:i+4])[0]     # The channel slopes for scaling, REAL(4)
        i+=4

    ColOff = [None]*NumOutChans
    for idx in range(0,NumOutChans):
        ColOff[idx] = struct.unpack('f',data[i:i+4])[0]     # The channel offsets for scaling, REAL(4)
        i+=4

    LenDesc = struct.unpack('i',data[i:i+4])[0]             # The number of characters in the description string, INT(4)
    i+=4
    try:
        DescStr = ''.join(struct.unpack(str('c'*LenDesc),data[i:i+LenDesc]))     # DescStr converted to ASCII
    except:
        DescStr = data[i:i+LenDesc].decode("utf-8").strip()
    i+=LenDesc

    ChanName = [None]*(NumOutChans+1)
    for idx in range(0,NumOutChans+1):
        try:
            ChanName[idx] = ''.join(''.join(struct.unpack('c'*LenName,data[i:i+LenName]))).strip()  # variable channel names
        except:
            ChanName[idx] = data[i:i+LenName].decode("utf-8").strip()
        i+=LenName

    ChanUnit = [None]*(NumOutChans+1)
    for idx in range(0,NumOutChans+1):
        try:
            ChanUnit[idx] = ''.join(''.join(struct.unpack('c'*LenUnit,data[i:i+LenUnit]))).strip()  # variable units
        except:
            ChanUnit[idx] = data[i:i+LenUnit].decode("utf-8").strip()
        i+=LenUnit

    #-------------------------        
    # get the channel time series
    #-------------------------
    nPts = NT*NumOutChans                                       # number of data points in the file   

    if FileID == 1:
        PackedTime = struct.unpack('i'*NT,data[i:i+4*NT])[0]    # read the time data
        i+=4*NT
    
    PackedData = [None]*nPts
    for idx in range(0,nPts):
        PackedData[idx] = struct.unpack('h',data[i:i+2])[0]     # read the channel data
        i+=2

    #-------------------------
    # Scale the packed binary to real data
    #-------------------------
    Channels = np.zeros((NT,NumOutChans+1))                     # output channels (including time in column 1)
    for it in range(0,NT):
        data_slice = PackedData[NumOutChans*(it):NumOutChans*(it+1)]
        for idx, (datai, ColOffi, ColScli) in enumerate(zip(data_slice, ColOff, ColScl)):
            Channels[it,idx+1] = (datai - ColOffi) / ColScli

    if FileID == 1:
        for idx, (PackedTimei, TimeOffi, TimeScli) in enumerate(zip(PackedTime, TimeOff, TimeScl)):
            Channels[idx,0] = (PackedTimei - TimeOffi) / TimeScli
    else:
        for idx in range(0,NT):
            Channels[idx,0] = TimeOut1 + TimeIncr*idx

    return Channels, ChanName, ChanUnit, FileID, DescStr

def ReadFASTtext(FileName):

    f = open(FileName, 'r')

    DescStr = []
    ln = f.readline()
    i = 0
    while ln != '':
        if 'Time' in ln:
            break
        DescStr.append(ln.strip())
        ln = f.readline()
        
    DescStr = ' '.join(DescStr).strip()
    ChanName = ln.split()
    ChanUnit = f.readline().split()
    Channels = np.loadtxt(f)
    f.close()

    return Channels, ChanName, ChanUnit, None, DescStr


if __name__ == "__main__":

    FileName = '0.outb'
    # FileName = 'temp/OpenFAST/09.out'
    
    data, meta = ReadFASToutFormat(FileName, OutFileFmt=2)

    import matplotlib.pyplot as plt
    plt.figure()
    xvar = 'Time'
    yvar = 'GenPwr'
    plt.plot(data[xvar], data['GenPwr'])
    plt.xlabel('%s, %s'%(xvar, meta['units'][xvar]))
    plt.ylabel('%s, %s'%(yvar, meta['units'][yvar]))
    plt.show()



import sys
from pyts.runInput.main import readInput, run, write
import time

import pyts
import pyts.runInput.main as ptsm
import pyts.io.input as ptsin
from pyts.base import tsGrid
from pyts.phaseModels.main import Rinker, Uniform
import numpy as np

fname = sys.argv[1]
outname = sys.argv[2]



def run1():
    config = readInput(fname)
    tm0 = time.time()
    tsdat = run(config)
    write(tsdat, config, outname)
    print('TurbSim exited normally, runtime was %g seconds' % (time.time() - tm0))


def run2():
    tsinput = ptsin.read(fname)
    tsinput['URef'] = 16.0
    tsr = ptsm.cfg2tsrun(tsinput)

    tsr.cohere = pyts.cohereModels.main.nwtc()
    tsr.stress = pyts.stressModels.main.uniform(0,0,0)
    rho = 0.5
    tmax = 50.0
    tsr.phase = Rinker(rho, np.pi)  ### pgraf turned it off for testing!
    cg = tsr.grid
    tsr.grid = tsGrid(center=cg.center, ny=cg.n_y, nz=cg.n_z,
                      height=cg.height, width=cg.width,
                      time_sec=tmax, dt=cg.dt)
    tm0 = time.time()
    tsdata = tsr()  ## actually runs turbsim
    ptsm.write(tsdata, tsinput, fname=outname+"2")
    print('TurbSim exited normally, runtime was %g seconds' % (time.time() - tm0))


run1()

run2()



#-------------------------------------------------------------------------
# Name:        SegIntersect.py
# Purpose: It Calculates Intersection of 2 segments in space. Also contains a simple 2-point distance calculator.
#
# Author:      RRD
#
# Created:     24/10/2012
# Copyright:   (c) rdamiani 2012
# Licence:     <your licence>
#-------------------------------------------------------------------------


import numpy as np

def frustum(Db, Dt, H):
    """This function returns a frustum's volume and center of mass, CM

    INPUT:
    Parameters
    ----------
    Db : float,        base diameter
    Dt : float,        top diameter
    H : float,         height

    OUTPUTs:
    -------
    vol : float,        volume
    cm : float,        geometric centroid relative to bottom (center of mass if uniform density)

    """
    vol = frustumVol(Db, Dt, H, diamFlag=True)
    cm  = frustumCG(Db, Dt, H, diamFlag=True)
    #vol = np.pi/12*H * (Db**2 + Dt**2 + Db * Dt)
    #cm = H/4 * (Db**2 + 3*Dt**2 + 2*Db*Dt) / (Db**2 + Dt**2 + Db*Dt)
    return vol, cm


def frustumVol(rb, rt, h, diamFlag=False):
    """This function returns a frustum's volume with radii or diameter inputs.

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    vol : float (scalar/vector), volume
    """
    if diamFlag:
        # Convert diameters to radii
        rb *= 0.5
        rt *= 0.5
    return ( np.pi * (h/3.0) * (rb*rb + rt*rt + rb*rt) )

def frustumCG(rb, rt, h, diamFlag=False):
    """This function returns a frustum's center of mass/gravity (centroid) with radii or diameter inputs.
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (centroid)
    """
    if diamFlag:
        # Convert diameters to radii
        rb *= 0.5
        rt *= 0.5
    return (0.25*h * (rb**2 + 2.*rb*rt + 3.*rt**2) / (rb**2 + rb*rt + rt**2))

def frustumIzz(rb, rt, h, diamFlag=False):
    """This function returns a frustum's mass-moment of inertia (divided by density) about the 
    central (axial) z-axis with radii or diameter inputs.
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """
    if diamFlag:
        # Convert diameters to radii
        rb *= 0.5
        rt *= 0.5
    # Integrate 2*pi*r*r^2 dr dz from r=0 to r(z), z=0 to h
    # Also equals 0.3*Vol * (rt**5.0 - rb**5.0) / (rt**3.0 - rb**3.0)
    # Also equals (0.1*np.pi*h * (rt**5.0 - rb**5.0) / (rt - rb) )
    return (0.1*np.pi*h * (rt**4.0 + rb*rt**3 + rb**2*rt**2 + rb**3*rt + rb**4.0))

def frustumIxx(rb, rt, h, diamFlag=False):
    """This function returns a frustum's mass-moment of inertia (divided by density) about the 
    transverse x/y-axis passing through the center of mass with radii or diameter inputs.
    NOTE: This is for a SOLID frustum, not a shell

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    Ixx=Iyy : float (scalar/vector),  Moment of inertia about x/y-axis through center of mass (principle axes)
    """
    if diamFlag:
        # Convert diameters to radii
        rb *= 0.5
        rt *= 0.5
    # Integrate pi*r(z)^4/4 + pi*r(z)^2*(z-z_cg)^2 dz from z=0 to h
    A = 0.5*frustumIzz(rb, rt, h)
    B = np.pi*h**3/80.0 * ( (rb**4 + 4.*rb**3*rt + 10.*rb**2*rt**2 + 4.*rb*rt**3 + rt**4) / (rb**2 + rb*rt + rt**2) )
    return (A + B)


def frustumShellVol(rb, rt, t, h, diamFlag=False):
    """This function returns a frustum shell's volume (for computing mass with density) with radii or diameter inputs.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    t  : float (scalar/vector),  thickness
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (ventroid)
    """
    if diamFlag:
        # Convert diameters to radii
        rb *= 0.5
        rt *= 0.5
    # Integrate 2*pi*r*dr*dz from r=ri(z) to ro(z), z=0 to h
    rb_o = rb
    rb_i = rb-t
    rt_o = rt
    rt_i = rt-t
    #( (np.pi*h/3.0) * ( (rb_o**2 + rb_o*rt_o + rt_o**2) - (rb_i**2 + rb_i*rt_i + rt_i**2) ) )
    return (frustumVol(rb_o, rt_o, h) - frustumVol(rb_i, rt_i, h))


def frustumShellCG(rb, rt, t, h, diamFlag=False):
    """This function returns a frustum's center of mass/gravity (centroid) with radii or diameter inputs.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    t  : float (scalar/vector),  thickness
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    cg : float (scalar/vector),  center of mass/gravity (ventroid)
    """
    if diamFlag:
        # Convert diameters to radii
        rb *= 0.5
        rt *= 0.5
    # Integrate 2*pi*r*z*dr*dz/V from r=ri(z) to ro(z), z=0 to h
    rb_o = rb
    rb_i = rb-t
    rt_o = rt
    rt_i = rt-t
    A = (rb_o**2 + 2.0*rb_o*rt_o + 3.0*rt_o**2) - (rb_i**2 + 2.0*rb_i*rt_i + 3.0*rt_i**2)
    B = (rb_o**2 + rb_o*rt_o + rt_o**2) - (rb_i**2 + rb_i*rt_i + rt_i**2)
    return (h * A / 4.0 / B)



def frustumShellIzz(rb, rt, t, h, diamFlag=False):
    """This function returns a frustum's mass-moment of inertia (divided by density) about the 
    central (axial) z-axis with radii or diameter inputs.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    t  : float (scalar/vector),  thickness
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    Izz : float (scalar/vector),  Moment of inertia about z-axis
    """
    if diamFlag:
        # Convert diameters to radii
        rb *= 0.5
        rt *= 0.5
    # Integrate 2*pi*r*dr*dz from r=ri(z) to ro(z), z=0 to h
    rb_o = rb
    rb_i = rb-t
    rt_o = rt
    rt_i = rt-t
    return (frustumIzz(rb_o, rt_o, h) - frustumIzz(rb_i, rt_i, h))



def frustumShellIxx(rb, rt, t, h, diamFlag=False):
    """This function returns a frustum's mass-moment of inertia (divided by density) about the 
    transverse x/y-axis passing through the center of mass with radii or diameter inputs.
    NOTE: This is for a frustum SHELL, not a solid

    INPUTS:
    Parameters
    ----------
    rb : float (scalar/vector),  base radius
    rt : float (scalar/vector),  top radius
    t  : float (scalar/vector),  thickness
    h  : float (scalar/vector),  height
    diamFlag : boolean, True if rb and rt are entered as diameters

    OUTPUTs:
    -------
    Ixx=Iyy : float (scalar/vector),  Moment of inertia about x/y-axis through center of mass (principle axes)
    """
    if diamFlag:
        # Convert diameters to radii
        rb *= 0.5
        rt *= 0.5
    # Integrate 2*pi*r*dr*dz from r=ri(z) to ro(z), z=0 to h
    rb_o = rb
    rb_i = rb-t
    rt_o = rt
    rt_i = rt-t
    return (frustumIxx(rb_o, rt_o, h) - frustumIxx(rb_i, rt_i, h))

if __name__ == '__main__':
    Db=6.5
    Dt=4.
    H=120.

    print ('From commonse.Frustum: Sample Volume and CM of FRUSTUM='+4*'{:8.4f}, ').format(*frustum(Db,Dt,H)[0].flatten())

def main():
    pass



from openmdao.api import ExplicitComponent
import numpy as np
from scipy.integrate import cumtrapz

from wisdem.commonse import gravity, eps, DirectionVector, NFREQ
from wisdem.commonse.utilities import assembleI, unassembleI
from .map_mooring import NLINES_MAX
        
class SubstructureGeometry(ExplicitComponent):
    """
    OpenMDAO Component class for substructure geometry for floating offshore wind turbines.
    """

    def initialize(self):
        self.options.declare('nFull')
        self.options.declare('nFullTow')
        
    def setup(self):
        nFull    = self.options['nFull']
        nFullTow = self.options['nFullTow']

        # Design variables
        self.add_input('main_d_full', val=np.zeros((nFull,)), units='m', desc='outer radius at each section node bottom to top (length = nsection + 1)')
        self.add_input('offset_d_full', val=np.zeros((nFull,)), units='m', desc='outer radius at each section node bottom to top (length = nsection + 1)')
        self.add_input('offset_z_nodes', val=np.zeros((nFull,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('offset_freeboard', val=0.0, units='m', desc='Length of column above water line')
        self.add_input('offset_draft', val=0.0, units='m', desc='Length of column below water line')
        self.add_input('main_z_nodes', val=np.zeros((nFull,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('fairlead_location', val=0.0, desc='Fractional length from column bottom to top for mooring line attachment')
        self.add_input('fairlead_offset_from_shell', val=0.0, units='m',desc='fairlead offset from shell')
        self.add_input('radius_to_offset_column', val=0.0, units='m',desc='Distance from main column centerpoint to offset column centerpoint')
        self.add_input('number_of_offset_columns', val=0, desc='Number of offset columns evenly spaced around main column')
        self.add_input('tower_d_full', val=np.zeros((nFullTow,)), units='m', desc='outer radius at each section node bottom to top (length = nsection + 1)')
        self.add_input('Rhub', val=0.0, units='m', desc='rotor hub radius')
        self.add_input('Hs', val=0.0, units='m', desc='significant wave height')
        self.add_input('max_survival_heel', val=0.0, units='deg', desc='max heel angle for turbine survival')
        
        # Output constraints
        self.add_output('fairlead', val=0.0, units='m', desc='Depth below water line for mooring line attachment')
        self.add_output('fairlead_radius', val=0.0, units='m', desc='Outer spar radius at fairlead depth (point of mooring attachment)')
        self.add_output('main_offset_spacing', val=0.0, desc='Radius of main and offset columns relative to spacing')
        self.add_output('tower_transition_buffer', val=0.0, units='m', desc='Buffer between substructure main and tower main')
        self.add_output('nacelle_transition_buffer', val=0.0, units='m', desc='Buffer between tower top and nacelle main')
        self.add_output('offset_freeboard_heel_margin', val=0.0, units='m', desc='Margin so offset column does not submerge during max heel')
        self.add_output('offset_draft_heel_margin', val=0.0, units='m', desc='Margin so offset column does not leave water during max heel')
        self.add_output('wave_height_fairlead_ratio', val=0.0, desc='Ratio of maximum wave height (avg of top 1%) to fairlead')

        
        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs):
        """Sets nodal points and sectional centers of mass in z-coordinate system with z=0 at the waterline.
        Nodal points are the beginning and end points of each section.
        Nodes and sections start at bottom and move upwards.
        
        INPUTS:
        ----------
        params   : dictionary of input parameters
        outputs : dictionary of output parameters
        
        OUTPUTS  : none (all unknown dictionary values set)
        """
        # Unpack variables
        ncolumns        = int(inputs['number_of_offset_columns'])
        R_od_main       = 0.5*inputs['main_d_full']
        R_od_offset    = 0.5*inputs['offset_d_full']
        R_semi          = inputs['radius_to_offset_column']
        R_tower         = 0.5*inputs['tower_d_full']
        R_hub           = inputs['Rhub']
        
        z_nodes_offset = inputs['offset_z_nodes']
        z_nodes_main    = inputs['main_z_nodes']
        
        location        = inputs['fairlead_location']
        fair_off        = inputs['fairlead_offset_from_shell']
        off_freeboard   = inputs['offset_freeboard']
        off_draft       = inputs['offset_draft']
        max_heel        = inputs['max_survival_heel']

        # Set spacing constraint
        outputs['main_offset_spacing'] = R_semi - R_od_main.max() - R_od_offset.max()

        # Determine location and radius at mooring connection point (fairlead)
        if ncolumns > 0:
            z_fairlead = location * (z_nodes_offset[-1] - z_nodes_offset[0]) + z_nodes_offset[0]
            outputs['fairlead_radius'] = R_semi + fair_off + np.interp(z_fairlead, z_nodes_offset, R_od_offset)
        else:
            z_fairlead = location * (z_nodes_main[-1] - z_nodes_main[0]) + z_nodes_main[0]
            outputs['fairlead_radius'] = fair_off + np.interp(z_fairlead, z_nodes_main, R_od_main)
        outputs['fairlead'] = -z_fairlead # Fairlead defined as positive below waterline
        outputs['wave_height_fairlead_ratio'] = inputs['Hs'] / np.abs(z_fairlead)
        
        # Constrain spar top to be at least greater than tower main
        outputs['tower_transition_buffer']   = R_od_main[-1] - R_tower[0]
        outputs['nacelle_transition_buffer'] = (R_hub + 1.0) - R_tower[-1] # Guessing at 6m size for nacelle

        # Make sure semi columns don't get submerged
        heel_deflect = R_semi*np.sin(np.deg2rad(max_heel))
        outputs['offset_freeboard_heel_margin'] = off_freeboard - heel_deflect
        outputs['offset_draft_heel_margin']     = off_draft - heel_deflect



class Substructure(ExplicitComponent):

    def initialize(self):
        self.options.declare('nFull')
        self.options.declare('nFullTow')
        
    def setup(self):
        nFull    = self.options['nFull']
        nFullTow = self.options['nFullTow']

        # Environment
        self.add_input('water_density', val=0.0, units='kg/m**3', desc='density of water')
        self.add_input('wave_period_range_low', val=2.0, units='s', desc='Lower bound of typical ocean wavve period')
        self.add_input('wave_period_range_high', val=20.0, units='s', desc='Upper bound of typical ocean wavve period')

        # From other components
        self.add_input('operational_heel', val=0.0, units='deg',desc='Maximum angle of heel allowable')
        self.add_input('mooring_mass', val=0.0, units='kg', desc='Mass of mooring lines')
        self.add_input('mooring_moments_of_inertia', val=np.zeros(6), units='kg*m**2', desc='mass moment of inertia of mooring system about fairlead-centerline point [xx yy zz xy xz yz]')
        self.add_input('mooring_neutral_load', val=np.zeros((NLINES_MAX,3)), units='N', desc='z-force of mooring lines on structure')
        self.add_input('mooring_surge_restoring_force', val=0.0, units='N', desc='Restoring force from mooring system after surge motion')
        self.add_input('mooring_pitch_restoring_force', val=np.zeros((NLINES_MAX,3)), units='N', desc='Restoring force from mooring system after pitch motion')
        self.add_input('mooring_cost', val=0.0, units='USD', desc='Cost of mooring system')
        self.add_input('mooring_stiffness', val=np.zeros((6,6)), units='N/m', desc='Linearized stiffness matrix of mooring system at neutral (no offset) conditions.')
        self.add_input('fairlead', val=1.0, units='m', desc='Depth below water for mooring line attachment')
        self.add_input('fairlead_radius', val=0.0, units='m', desc='Outer spar radius at fairlead depth (point of mooring attachment)')
        
        self.add_input('number_of_offset_columns', val=0, desc='Number of offset columns evenly spaced around main column')
        self.add_input('radius_to_offset_column', val=0.0, units='m',desc='Distance from main column centerpoint to offset column centerpoint')

        self.add_input('main_Iwaterplane', val=0.0, units='m**4', desc='Second moment of area of waterplane cross-section')
        self.add_input('main_Awaterplane', val=0.0, units='m**2', desc='Area of waterplane cross-section')
        self.add_input('main_cost', val=0.0, units='USD', desc='Cost of spar structure')
        self.add_input('main_mass', val=np.zeros((nFull-1,)), units='kg', desc='mass of main column by section')
        self.add_input('main_freeboard', val=0.0, units='m', desc='Length of spar above water line')
        self.add_input('main_center_of_buoyancy', val=0.0, units='m', desc='z-position of center of column buoyancy force')
        self.add_input('main_center_of_mass', val=0.0, units='m', desc='z-position of center of column mass')
        self.add_input('main_moments_of_inertia', val=np.zeros(6), units='kg*m**2', desc='mass moment of inertia of column about main [xx yy zz xy xz yz]')
        self.add_input('main_added_mass', val=np.zeros(6), units='kg', desc='Diagonal of added mass matrix- masses are first 3 entries, moments are last 3')

        self.add_input('offset_Iwaterplane', val=0.0, units='m**4', desc='Second moment of area of waterplane cross-section')
        self.add_input('offset_Awaterplane', val=0.0, units='m**2', desc='Area of waterplane cross-section')
        self.add_input('offset_cost', val=0.0, units='USD', desc='Cost of spar structure')
        self.add_input('offset_mass', val=np.zeros((nFull-1,)), units='kg', desc='mass of offset column by section')
        self.add_input('offset_center_of_buoyancy', val=0.0, units='m', desc='z-position of center of column buoyancy force')
        self.add_input('offset_center_of_mass', val=0.0, units='m', desc='z-position of center of column mass')
        self.add_input('offset_moments_of_inertia', val=np.zeros(6), units='kg*m**2', desc='mass moment of inertia of column about main [xx yy zz xy xz yz]')
        self.add_input('offset_added_mass', val=np.zeros(6), units='kg', desc='Diagonal of added mass matrix- masses are first 3 entries, moments are last 3')

        self.add_input('tower_mass', val=0.0, units='kg', desc='Mass of tower')
        self.add_input('tower_shell_cost', val=0.0, units='USD', desc='Cost of tower')
        self.add_input('tower_I_base', val=np.zeros(6), units='kg*m**2', desc='Moments about tower main')
        self.add_input('tower_z_full', val=np.zeros((nFullTow,)), units='m', desc='z-coordinates of section nodes (length = nsection+1)')
        self.add_input('rna_mass', val=0.0, units='kg', desc='Mass of RNA')
        self.add_input('rna_cg', val=np.zeros(3), units='m', desc='Location of RNA center of mass relative to tower top')
        self.add_input('rna_I', val=np.zeros(6), units='kg*m**2', desc='Moments about turbine main')
        
        self.add_input('water_ballast_zpts_vector', val=np.zeros((nFull,)), units='m', desc='z-points of potential ballast mass')
        self.add_input('water_ballast_radius_vector', val=np.zeros((nFull,)), units='m', desc='Inner radius of potential ballast mass')

        self.add_input('structural_mass', val=0.0, units='kg', desc='Mass of whole turbine except for mooring lines')
        self.add_input('structure_center_of_mass', val=np.zeros(3), units='m', desc='xyz-position of center of gravity of whole turbine')
        self.add_input('structural_frequencies', val=np.zeros(NFREQ), units='Hz', desc='')
        self.add_input('z_center_of_buoyancy', val=0.0, units='m', desc='z-position of center of gravity (x,y = 0,0)')
        self.add_input('total_displacement', val=0.0, units='m**3', desc='Total volume of water displaced by floating turbine (except for mooring lines)')
        self.add_input('total_force', val=np.zeros(3), units='N', desc='Net forces on turbine')
        self.add_input('total_moment', val=np.zeros(3), units='N*m', desc='Moments on whole turbine')

        self.add_input('pontoon_cost', val=0.0, units='USD', desc='Cost of pontoon elements and connecting truss')

        
        # Outputs
        self.add_output('substructure_moments_of_inertia', val=np.zeros(6), units='kg*m**2', desc='mass moment of inertia of substructure (no tower or rna or mooring) [xx yy zz xy xz yz]')
        self.add_output('total_mass', val=0.0, units='kg', desc='total mass of spar and moorings')
        self.add_output('total_cost', val=0.0, units='USD', desc='total cost of spar and moorings')
        self.add_output('metacentric_height', val=0.0, units='m', desc='measure of static overturning stability')
        self.add_output('buoyancy_to_gravity', val=0.0, desc='static stability margin based on position of centers of gravity and buoyancy')
        self.add_output('offset_force_ratio', val=0.0, desc='total surge force divided by restoring force')
        self.add_output('heel_moment_ratio', val=0.0, desc='total pitch moment divided by restoring moment')
        self.add_output('Iwaterplane_system', val=0.0, units='m**4', desc='Second moment of area of waterplane cross-section for whole structure')

        self.add_output('center_of_mass', val=np.zeros(3), units='m', desc='xyz-position of center of gravity (x,y = 0,0)')

        self.add_output('variable_ballast_mass', val=0.0, units='kg', desc='Amount of variable water ballast')
        self.add_output('variable_ballast_center_of_mass', val=0.0, units='m', desc='Center of mass for variable ballast')
        self.add_output('variable_ballast_moments_of_inertia', val=np.zeros(6), units='kg*m**2', desc='mass moment of inertia of variable ballast [xx yy zz xy xz yz]')
        self.add_output('variable_ballast_height', val=0.0, units='m', desc='height of water ballast to balance spar')
        self.add_output('variable_ballast_height_ratio', val=0.0, desc='Ratio of water ballast height to available height')

        self.add_output('mass_matrix', val=np.zeros(6), units='kg', desc='Summary mass matrix of structure (minus pontoons)')
        self.add_output('added_mass_matrix', val=np.zeros(6), units='kg', desc='Summary hydrodynamic added mass matrix of structure (minus pontoons)')
        self.add_output('hydrostatic_stiffness', val=np.zeros(6), units='N/m', desc='Summary hydrostatic stiffness of structure')
        self.add_output('rigid_body_periods', val=np.zeros(6), units='s', desc='Natural periods of oscillation in 6 DOF')
        self.add_output('period_margin_low', val=np.zeros(6), desc='Margin between natural periods and 2 second wave period')
        self.add_output('period_margin_high', val=np.zeros(6), desc='Margin between natural periods and 20 second wave period')
        self.add_output('modal_margin_low', val=np.zeros(NFREQ), desc='Margin between structural modes and 2 second wave period')
        self.add_output('modal_margin_high', val=np.zeros(NFREQ), desc='Margin between structural modes and 20 second wave period')
        
        
        # Derivatives
        self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs):
        # TODO: Get centerlines right- in sparGeometry?
        # Determine ballast and cg of system
        self.balance(inputs, outputs)
        
        # Determine stability, metacentric height from waterplane profile, displaced volume
        self.compute_stability(inputs, outputs)

        # Compute natural periods of osciallation
        self.compute_rigid_body_periods(inputs, outputs)
        
        # Check margins of natural and eigenfrequencies against waves
        self.check_frequency_margins(inputs, outputs)
        
        # Sum all costs
        self.compute_costs(inputs, outputs)

        
    def balance(self, inputs, outputs):
        # Unpack variables
        m_struct     = inputs['structural_mass']
        Fz_mooring   = np.sum( inputs['mooring_neutral_load'][:,-1] )
        m_mooring    = inputs['mooring_mass']
        
        V_system     = inputs['total_displacement']

        cg_struct    = inputs['structure_center_of_mass']
        
        z_water_data = inputs['water_ballast_zpts_vector']
        r_water_data = inputs['water_ballast_radius_vector']
        rhoWater     = inputs['water_density']
        
        # SEMI TODO: Make water_ballast in main only?  columns too?  How to apportion?

        # Make sure total mass of system with variable water ballast balances against displaced volume
        # Water ballast should be buried in m_column
        m_water  = V_system*rhoWater - (m_struct + Fz_mooring/gravity)
        m_system = m_struct + m_water

        # Output substructure total turbine mass
        outputs['total_mass'] = m_struct + m_mooring

        # Find height given interpolant functions from columns
        m_water_data = rhoWater * np.pi * cumtrapz(r_water_data**2, z_water_data)
        m_water_data = np.r_[0.0, m_water_data] #cumtrapz has length-1
        
        if m_water_data[-1] < m_water:
            # Don't have enough space, so max out variable balast here and constraints will catch this
            z_end = z_water_data[-1]
            coeff = m_water / m_water_data[-1]
        elif m_water < 0.0:
            z_end = z_water_data[0]
            coeff = 0.0
        else:
            z_end = np.interp(m_water, m_water_data, z_water_data)
            coeff = 1.0
        h_water = z_end - z_water_data[0]
        outputs['variable_ballast_mass']   = m_water
        outputs['variable_ballast_height'] = coeff * h_water
        outputs['variable_ballast_height_ratio'] = coeff * h_water / (z_water_data[-1] - z_water_data[0])
        
        # Find cg of whole system
        # First find cg of water variable ballast by finding midpoint of mass sum
        z_cg  = np.interp(0.5*coeff*m_water, m_water_data, z_water_data)
        outputs['center_of_mass'] = (m_struct*cg_struct + m_water*np.r_[0.0, 0.0, z_cg]) / m_system
        outputs['variable_ballast_center_of_mass'] = z_cg

        # Integrate for moment of inertia of variable ballast
        npts  = 100
        z_int = np.linspace(float(z_water_data[0]), float(z_end), npts)
        r_int = np.interp(z_int, z_water_data, r_water_data)
        Izz   = 0.5 * rhoWater * np.pi * np.trapz(r_int**4, z_int)
        Ixx   = rhoWater * np.pi * np.trapz(0.25*r_int**4 + r_int**2*(z_int-z_cg)**2, z_int)
        outputs['variable_ballast_moments_of_inertia'] = np.array([Ixx, Ixx, Izz, 0.0, 0.0, 0.0])

        
    def compute_stability(self, inputs, outputs):
        # Unpack variables
        ncolumn         = int(inputs['number_of_offset_columns'])
        z_cb            = inputs['z_center_of_buoyancy']
        z_cg            = outputs['center_of_mass'][-1]
        V_system        = inputs['total_displacement']
        
        Iwater_main     = inputs['main_Iwaterplane']
        Iwater_column   = inputs['offset_Iwaterplane']
        Awater_column   = inputs['offset_Awaterplane']

        F_surge         = inputs['total_force'][0]
        M_pitch         = inputs['total_moment'][1]
        F_restore       = inputs['mooring_surge_restoring_force']
        rhoWater        = inputs['water_density']
        R_semi          = inputs['radius_to_offset_column']

        F_restore_pitch = inputs['mooring_pitch_restoring_force']
        z_fairlead      = inputs['fairlead']*(-1)
        R_fairlead      = inputs['fairlead_radius']
        oper_heel       = float(inputs['operational_heel'])
        
        # Compute the distance from the center of buoyancy to the metacentre (BM is naval architecture)
        # BM = Iw / V where V is the displacement volume (just computed)
        # Iw is the area moment of inertia (meters^4) of the water-plane cross section about the heel axis
        # For a spar, we assume this is just the I of a circle about x or y
        # See https://en.wikipedia.org/wiki/Metacentric_height
        # https://en.wikipedia.org/wiki/List_of_second_moments_of_area
        # and http://farside.ph.utexas.edu/teaching/336L/Fluidhtml/node30.html

        # Water plane area of all components with parallel axis theorem
        Iwater_system = Iwater_main
        radii = R_semi * np.cos( np.linspace(0, 2*np.pi, ncolumn+1) )
        for k in range(ncolumn):
            Iwater_system += Iwater_column + Awater_column*radii[k]**2
        outputs['Iwaterplane_system'] = Iwater_column
            
        # Measure static stability:
        # 1. Center of buoyancy should be above CG (difference should be positive)
        # 2. Metacentric height should be positive
        buoyancy2metacentre_BM          = Iwater_system / V_system
        outputs['buoyancy_to_gravity'] = z_cg - z_cb
        outputs['metacentric_height' ] = buoyancy2metacentre_BM - outputs['buoyancy_to_gravity']

        F_buoy     = V_system * rhoWater * gravity
        M_restore  = outputs['metacentric_height'] * np.sin(np.deg2rad(oper_heel)) * F_buoy 

        # Convert mooring restoring force after pitch to a restoring moment
        nlines = np.count_nonzero(F_restore_pitch[:,2])
        F_restore_pitch = F_restore_pitch[:nlines,:]
        moorx  = R_fairlead * np.cos( np.linspace(0, 2*np.pi, nlines+1)[:-1] )
        moory  = R_fairlead * np.sin( np.linspace(0, 2*np.pi, nlines+1)[:-1] )
        r_moor = np.c_[moorx, moory, (z_fairlead - z_cg)*np.ones(moorx.shape)]
        Msum   = 0.0
        for k in range(nlines):
            dvF   = DirectionVector.fromArray(F_restore_pitch[k,:])
            dvR   = DirectionVector.fromArray(r_moor[k,:]).yawToHub(oper_heel)
            M     = dvR.cross(dvF)
            Msum += M.y

        M_restore += Msum
        
        # Comput heel angle, scaling overturning moment by defect of inflow velocity
        # TODO: Make this another load case in Frame3DD
        outputs['heel_moment_ratio'] =  np.abs( np.cos(np.deg2rad(oper_heel))**2.0 * M_pitch / M_restore )

        # Now compute offsets from the applied force
        # First use added mass (the mass of the water that must be displaced in movement)
        # http://www.iaea.org/inis/collection/NCLCollectionStore/_Public/09/411/9411273.pdf
        #mass_add_surge = rhoWater * np.pi * R_od.max() * draft
        #T_surge        = 2*np.pi*np.sqrt( (outputs['total_mass']+mass_add_surge) / kstiff_horiz_mooring)

        # Compare restoring force from mooring to force of worst case spar displacement
        outputs['offset_force_ratio'] = np.abs(F_surge / F_restore)


    def compute_rigid_body_periods(self, inputs, outputs):
        # Unpack variables
        ncolumn         = int(inputs['number_of_offset_columns'])
        R_semi          = inputs['radius_to_offset_column']
        
        m_main          = np.sum(inputs['main_mass'])
        m_column        = np.sum(inputs['offset_mass'])
        m_tower         = np.sum(inputs['tower_mass'])
        m_rna           = inputs['rna_mass']
        m_mooring       = inputs['mooring_mass']
        m_total         = outputs['total_mass']
        m_water         = np.maximum(0.0, outputs['variable_ballast_mass'])
        m_a_main        = inputs['main_added_mass']
        m_a_column      = inputs['offset_added_mass']
        
        rhoWater        = inputs['water_density']
        V_system        = inputs['total_displacement']
        h_metacenter    = outputs['metacentric_height']

        Awater_main     = inputs['main_Awaterplane']
        Awater_column   = inputs['offset_Awaterplane']
        I_main          = inputs['main_moments_of_inertia']
        I_column        = inputs['offset_moments_of_inertia']
        I_mooring       = inputs['mooring_moments_of_inertia']
        I_water         = outputs['variable_ballast_moments_of_inertia']
        I_tower         = inputs['tower_I_base']
        I_rna           = inputs['rna_I']
        I_waterplane    = outputs['Iwaterplane_system']

        z_cg_main       = float(inputs['main_center_of_mass'])
        z_cb_main       = float(inputs['main_center_of_buoyancy'])
        z_cg_column     = float(inputs['offset_center_of_mass'])
        z_cb_column     = float(inputs['offset_center_of_buoyancy'])
        z_cb            = float(inputs['z_center_of_buoyancy'])
        z_cg_water      = float(outputs['variable_ballast_center_of_mass'])
        z_fairlead      = float(inputs['fairlead']*(-1))
        
        r_cg            = outputs['center_of_mass']
        cg_rna          = inputs['rna_cg']
        z_tower         = inputs['tower_z_full']
        
        K_moor          = np.diag( inputs['mooring_stiffness'] )

        
        # Number of degrees of freedom
        nDOF = 6

        # Compute elements on mass matrix diagonal
        M_mat = np.zeros((nDOF,))
        # Surge, sway, heave just use normal inertia (without mooring according to Senu)
        M_mat[:3] = m_total + m_water - m_mooring
        # Add in moments of inertia of primary column
        I_total = assembleI( np.zeros(6) )
        I_main  = assembleI( I_main )
        R       = np.array([0.0, 0.0, z_cg_main]) - r_cg
        I_total += I_main + m_main*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        # Add up moments of intertia of other columns
        radii_x   = R_semi * np.cos( np.linspace(0, 2*np.pi, ncolumn+1) )
        radii_y   = R_semi * np.sin( np.linspace(0, 2*np.pi, ncolumn+1) )
        I_column  = assembleI( I_column )
        for k in range(ncolumn):
            R        = np.array([radii_x[k], radii_y[k], z_cg_column]) - r_cg
            I_total += I_column + m_column*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        # Add in variable ballast
        R         = np.array([0.0, 0.0, z_cg_water]) - r_cg
        I_total  += assembleI(I_water) + m_water*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        
        # Save what we have so far as m_substructure & I_substructure and move to its own CM
        m_subs    =  m_main           + ncolumn*m_column             + m_water
        z_cg_subs = (m_main*z_cg_main + ncolumn*m_column*z_cg_column + m_water*z_cg_water) / m_subs
        R              = r_cg - np.array([0.0, 0.0, z_cg_subs])
        I_substructure = I_total + m_subs*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        outputs['substructure_moments_of_inertia'] = unassembleI( I_total )

        # Now go back to the total
        # Add in mooring system- Not needed according to Senu
        #R         = np.array([0.0, 0.0, z_fairlead]) - r_cg
        #I_total  += assembleI(I_mooring) + m_mooring*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        # Add in tower
        R         = np.array([0.0, 0.0, z_tower[0]]) - r_cg
        I_total  += assembleI(I_tower) + m_tower*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        # Add in RNA
        R         = np.array([0.0, 0.0, z_tower[-1]]) + cg_rna - r_cg
        I_total  += assembleI(I_rna) + m_rna*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        # Stuff moments of inertia into mass matrix
        M_mat[3:] = unassembleI( I_total )[:3]
        outputs['mass_matrix'] = M_mat
        
        # Add up all added mass entries in a similar way
        A_mat = np.zeros((nDOF,))
        # Surge, sway, heave just use normal inertia
        A_mat[:3] = m_a_main[:3] + ncolumn*m_a_column[:3]
        # Add up moments of inertia, move added mass moments from CofB to CofG
        I_main    = assembleI( np.r_[m_a_main[3:]  , np.zeros(3)] )
        R         = np.array([0.0, 0.0, z_cb_main]) - r_cg
        I_total   = I_main + m_a_main[0]*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        # Add up added moments of intertia of all columns for other entries
        I_column  = assembleI( np.r_[m_a_column[3:], np.zeros(3)] )
        for k in range(ncolumn):
            R        = np.array([radii_x[k], radii_y[k], z_cb_column]) - r_cg
            I_total += I_column + m_a_column[0]*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        A_mat[3:] = unassembleI( I_total )[:3]
        outputs['added_mass_matrix'] = A_mat
        
        # Hydrostatic stiffness has contributions in heave (K33) and roll/pitch (K44/55)
        # See DNV-RP-H103: Modeling and Analyis of Marine Operations
        K_hydro = np.zeros((nDOF,))
        K_hydro[2]   = rhoWater * gravity * (Awater_main + ncolumn*Awater_column)
        K_hydro[3:5] = rhoWater * gravity * V_system * h_metacenter # FAST eqns: (I_waterplane + V_system * z_cb)
        outputs['hydrostatic_stiffness'] = K_hydro

        # Now compute all six natural periods at once
        epsilon = 1e-6 # Avoids numerical issues
        K_total = np.maximum(K_hydro + K_moor, 0.0)
        outputs['rigid_body_periods'] = 2*np.pi * np.sqrt( (M_mat + A_mat) / (K_total + epsilon) )

        
    def check_frequency_margins(self, inputs, outputs):
        # Unpack variables
        T_sys       = outputs['rigid_body_periods']
        T_wave_low  = inputs['wave_period_range_low']
        T_wave_high = inputs['wave_period_range_high']
        f_struct    = inputs['structural_frequencies']
        T_struct    = 1.0 / f_struct

        # Waves cannot excite yaw, so removing that constraint
        
        # Compute margins between wave forcing and natural periods
        indicator_high = T_wave_high * np.ones(T_sys.shape)
        indicator_high[T_sys < T_wave_low] = 1e-16
        indicator_high[-1] = 1e-16 # Not yaw
        outputs['period_margin_high'] = T_sys / indicator_high

        indicator_low = T_wave_low * np.ones(T_sys.shape)
        indicator_low[T_sys > T_wave_high] = 1e30
        indicator_low[-1] = 1e30 # Not yaw
        outputs['period_margin_low']  = T_sys / indicator_low

        # Compute margins bewteen wave forcing and structural frequencies
        indicator_high = T_wave_high * np.ones(T_struct.shape)
        indicator_high[T_struct < T_wave_low] = 1e-16
        outputs['modal_margin_high'] = T_struct / indicator_high

        indicator_low = T_wave_low * np.ones(T_struct.shape)
        indicator_low[T_struct > T_wave_high] = 1e30
        outputs['modal_margin_low']  = T_struct / indicator_low
        
        
    def compute_costs(self, inputs, outputs):
        # Unpack variables
        ncolumn    = int(inputs['number_of_offset_columns'])
        c_mooring  = inputs['mooring_cost']
        c_aux      = inputs['offset_cost']
        c_main     = inputs['main_cost']
        c_pontoon  = inputs['pontoon_cost']
        c_tower    = inputs['tower_shell_cost']

        outputs['total_cost'] = c_mooring + ncolumn*c_aux + c_main + c_pontoon + c_tower
        




#!/usr/bin/env python
# encoding: utf-8
"""
towerstruc.py

Created by Andrew Ning on 2012-01-20.
Copyright (c) NREL. All rights reserved.

HISTORY:  2012 created
          -7/2014:  R.D. Bugs found in the call to shellBucklingEurocode from towerwithFrame3DD. Fixed.
                    Also set_as_top added.
          -10/2014: R.D. Merged back with some changes Andrew did on his end.
          -12/2014: A.N. fixed some errors from the merge (redundant drag calc).  pep8 compliance.  removed several unneccesary variables and imports (including set_as_top)
          - 6/2015: A.N. major rewrite.  removed pBEAM.  can add spring stiffness anywhere.  can add mass anywhere.
            can use different material props throughout.
          - 7/2015 : R.D. modified to use commonse modules.
          - 1/2018 : G.B. modified for easier use with other modules, reducing user input burden, and shifting more to commonse
 """

from __future__ import print_function

import numpy as np
from openmdao.api import ExplicitComponent, Group, Problem, IndepVarComp

from wisdem.commonse.WindWaveDrag import AeroHydroLoads, CylinderWindDrag, CylinderWaveDrag

from wisdem.commonse.environment import WindBase, WaveBase, LinearWaves, TowerSoil, PowerWind, LogWind
from wisdem.commonse.tube import CylindricalShellProperties
from wisdem.commonse.utilities import assembleI, unassembleI, nodal2sectional, interp_with_deriv
from wisdem.commonse import gravity, eps, NFREQ

from wisdem.commonse.vertical_cylinder import CylinderDiscretization, CylinderMass, CylinderFrame3DD

import wisdem.commonse.UtilizationSupplement as Util


def find_nearest(array,value):
    return (np.abs(array-value)).argmin() 


# -----------------
#  Components
# -----------------

class MonopileFoundation(ExplicitComponent):
    def initialize(self):
        self.options.declare('nPoints')
        self.options.declare('monopile')
    
    def setup(self):
        nPoints = self.options['nPoints']

        self.add_input('tower_section_height', np.zeros(nPoints-1), units='m', desc='parameterized section heights along cylinder')
        self.add_input('tower_outer_diameter', np.zeros(nPoints), units='m', desc='cylinder diameter at corresponding locations')
        self.add_input('tower_wall_thickness', np.zeros(nPoints-1), units='m', desc='shell thickness at corresponding locations')
        self.add_input('suctionpile_depth', 0.0, units='m', desc='depth of foundation in the soil')
        self.add_input('suctionpile_depth_diam_ratio', 0.0, desc='ratio of sunction pile depth to mudline monopile diameter')
        self.add_input('foundation_height', 0.0, units='m', desc='height of foundation (0.0 for land, -water_depth for fixed bottom)')

        nadd = 1 if self.options['monopile'] else 0
        self.add_output('section_height_out', np.zeros(nPoints-1+nadd), units='m', desc='parameterized section heights along cylinder')
        self.add_output('outer_diameter_out', np.zeros(nPoints+nadd), units='m', desc='cylinder diameter at corresponding locations')
        self.add_output('wall_thickness_out', np.zeros(nPoints-1+nadd), units='m', desc='shell thickness at corresponding locations')
        self.add_output('foundation_height_out', 0.0, units='m', desc='height of suction pile bottom (0.0 for land, -water_depth-pile for fixed bottom)')

    def compute(self, inputs, outputs):
        if self.options['monopile']:
            # Determine suction pile depth
            pile = inputs['suctionpile_depth']
            if pile == 0.0:
                pile = inputs['suctionpile_depth_diam_ratio'] * inputs['tower_outer_diameter'][0]

            # Gravity foundations will have 0 suction depth, but we will still add an extra point for consistency
            pile = np.maximum(0.1, pile)
            
            outputs['section_height_out'] = np.r_[pile, inputs['tower_section_height']]
            outputs['outer_diameter_out'] = np.r_[inputs['tower_outer_diameter'][0], inputs['tower_outer_diameter']]
            outputs['wall_thickness_out'] = np.r_[inputs['tower_wall_thickness'][0], inputs['tower_wall_thickness']]
            outputs['foundation_height_out'] = inputs['foundation_height'] - pile
        else:
            outputs['section_height_out'] = inputs['tower_section_height']
            outputs['outer_diameter_out'] = inputs['tower_outer_diameter']
            outputs['wall_thickness_out'] = inputs['tower_wall_thickness']
            outputs['foundation_height_out'] = inputs['foundation_height']
            

class TowerDiscretization(ExplicitComponent):
    def initialize(self):
        self.options.declare('nPoints')
    
    def setup(self):
        nPoints = self.options['nPoints']

        self.add_input('hub_height', val=0.0, units='m', desc='diameter at tower base')
        self.add_input('z_param', np.zeros(nPoints), units='m', desc='parameterized locations along tower, linear lofting between')

        self.add_output('height_constraint', val=0.0, units='m', desc='mismatch between tower height and desired hub_height')

        self.declare_partials('height_constraint', ['hub_height','z_param'])
        
    def compute(self, inputs, outputs):
        outputs['height_constraint'] = inputs['hub_height'] - inputs['z_param'][-1]

    def compute_partials(self, inputs, J):
        nPoints = self.options['nPoints']
        
        J['height_constraint','hub_height'] = 1.
        J['height_constraint','z_param'] = np.zeros(nPoints)
        J['height_constraint','z_param'][-1] = -1.
        
        
        
class TowerMass(ExplicitComponent):

    def initialize(self):
        self.options.declare('nFull')
    
    def setup(self):
        nFull = self.options['nFull']
        
        self.add_input('cylinder_mass', val=np.zeros(nFull-1), units='kg', desc='Total cylinder mass')
        self.add_input('cylinder_cost', val=0.0, units='USD', desc='Total cylinder cost')
        self.add_input('cylinder_center_of_mass', val=0.0, units='m', desc='z-position of center of mass of cylinder')
        self.add_input('cylinder_section_center_of_mass', val=np.zeros(nFull-1), units='m', desc='z position of center of mass of each can in the cylinder')
        self.add_input('cylinder_I_base', np.zeros((6,)), units='kg*m**2', desc='mass moment of inertia of cylinder about base [xx yy zz xy xz yz]')

        self.add_input('transition_piece_height', 0.0, units='m', desc='point mass height of transition piece above water line')
        self.add_input('transition_piece_mass', 0.0, units='kg', desc='point mass of transition piece')
        self.add_input('gravity_foundation_mass', 0.0, units='kg', desc='extra mass of gravity foundation')
        self.add_input('foundation_height', 0.0, units='m', desc='height of foundation (0.0 for land, -water_depth for fixed bottom)')
        self.add_input('z_full', np.zeros(nFull), units='m', desc='parameterized locations along tower, linear lofting between')
        
        self.add_output('tower_raw_cost', val=0.0, units='USD', desc='Total tower cost')
        self.add_output('tower_mass', val=0.0, units='kg', desc='Total tower mass')
        self.add_output('tower_center_of_mass', val=0.0, units='m', desc='z-position of center of mass of tower')
        self.add_output('tower_section_center_of_mass', val=np.zeros(nFull-1), units='m', desc='z position of center of mass of each can in the tower')
        self.add_output('tower_I_base', np.zeros((6,)), units='kg*m**2', desc='mass moment of inertia of tower about base [xx yy zz xy xz yz]')

        self.add_output('monopile_mass', val=0.0, units='kg', desc='Mass of monopile from bottom of suction pile through transition piece')
        self.add_output('monopile_cost', val=0.0, units='USD', desc='Total monopile cost')
        self.add_output('monopile_length', val=0.0, units='m', desc='Length of monopile from bottom of suction pile through transition piece')
        
        self.declare_partials('tower_raw_cost', 'cylinder_cost')
        self.declare_partials('tower_mass', ['cylinder_mass','transition_piece_mass'])
        self.declare_partials('tower_center_of_mass', 'cylinder_center_of_mass')
        self.declare_partials('tower_section_center_of_mass', 'cylinder_section_center_of_mass')
        self.declare_partials('tower_I_base', 'cylinder_I_base')
        self.declare_partials('monopile_mass', ['cylinder_mass','z_full','transition_piece_height'])
        self.declare_partials('monopile_cost', ['cylinder_mass','z_full','transition_piece_height','cylinder_cost'])
        self.declare_partials('monopile_length', ['transition_piece_height','z_full'])

        self.J = {}
        
        
    def compute(self, inputs, outputs):
        outputs['tower_raw_cost']       = inputs['cylinder_cost']
        outputs['tower_mass']           = inputs['cylinder_mass'].sum()
        outputs['tower_center_of_mass'] = ( (inputs['cylinder_center_of_mass']*outputs['tower_mass'] +
                                             inputs['transition_piece_mass']*inputs['transition_piece_height'] +
                                             inputs['gravity_foundation_mass']*inputs['foundation_height']) /
                                            (outputs['tower_mass']+inputs['transition_piece_mass']+inputs['gravity_foundation_mass']) )
        outputs['tower_section_center_of_mass'] = inputs['cylinder_section_center_of_mass']
        outputs['tower_I_base']         = inputs['cylinder_I_base']

        outputs['monopile_mass'],dydx,dydxp,dydyp = interp_with_deriv(inputs['transition_piece_height'],
                                                                      inputs['z_full'],
                                                                      np.r_[0.0, np.cumsum(inputs['cylinder_mass'])])
        outputs['tower_mass']     -= outputs['monopile_mass']
        outputs['monopile_cost']   = inputs['cylinder_cost']*outputs['monopile_mass']/inputs['cylinder_mass'].sum()
        outputs['monopile_mass']  += inputs['transition_piece_mass'] + inputs['gravity_foundation_mass']
        outputs['monopile_length'] = inputs['transition_piece_height'] - inputs['z_full'][0]

        self.J = {}
        self.J['monopile_mass', 'z_full'] = dydxp[0,:]
        self.J['monopile_mass', 'cylinder_mass'] = dydyp[0,1:]
        self.J['monopile_mass', 'transition_piece_height'] = dydx[0,0]

        self.J['monopile_cost', 'z_full'] = inputs['cylinder_cost'] * self.J['monopile_mass', 'z_full'] / inputs['cylinder_mass'].sum()
        self.J['monopile_cost', 'cylinder_cost'] = outputs['monopile_mass']/ inputs['cylinder_mass']
        self.J['monopile_cost', 'cylinder_mass'] = inputs['cylinder_cost']*self.J['monopile_mass', 'cylinder_mass']/inputs['cylinder_mass'] - outputs['monopile_cost']/inputs['cylinder_mass']
        self.J['monopile_cost', 'transition_piece_height'] = inputs['cylinder_cost'] * self.J['monopile_mass', 'transition_piece_height'] / inputs['cylinder_mass']
        
        
    def compute_partials(self, inputs, J):
        J['tower_mass','cylinder_mass'] = np.ones(len(inputs['cylinder_mass'])) - self.J['monopile_mass', 'cylinder_mass']
        J['tower_mass','transition_piece_mass'] = 1.0
        J['tower_mass', 'z_full'] = -self.J['monopile_mass', 'z_full']
        J['tower_mass', 'transition_piece_height'] = -self.J['monopile_mass', 'transition_piece_height']

        J['tower_raw_cost','cylinder_cost'] = 1.0

        J['tower_center_of_mass','cylinder_center_of_mass'] = 1.0

        J['tower_section_center_of_mass','cylinder_section_center_of_mass'] = np.eye(len(inputs['cylinder_section_center_of_mass']))

        J['tower_I_base','cylinder_I_base'] = np.eye(len(inputs['cylinder_I_base']))

        J['monopile_mass', 'z_full'] = self.J['monopile_mass', 'z_full']
        J['monopile_mass', 'cylinder_mass'] = self.J['monopile_mass', 'cylinder_mass']
        J['monopile_mass', 'transition_piece_height'] = self.J['monopile_mass', 'transition_piece_height']
        J['monopile_mass', 'transition_piece_mass'] = 1.0
        J['monopile_mass', 'gravity_foundation_mass'] = 1.0

        J['monopile_cost', 'z_full'] = self.J['monopile_cost', 'z_full']
        J['monopile_cost', 'cylinder_cost'] = self.J['monopile_cost', 'cylinder_cost']
        J['monopile_cost', 'cylinder_mass'] = self.J['monopile_cost', 'cylinder_mass']
        J['monopile_cost', 'transition_piece_height'] = self.J['monopile_cost', 'transition_piece_height']

        J['monopile_length','transition_piece_height'] = 1.
        J['monopile_length','z_full'] = np.zeros(inputs['z_full'].size)
        J['monopile_length','z_full'][0] = -1.
        
        
        
class TurbineMass(ExplicitComponent):

    def setup(self):
        
        self.add_input('hub_height', val=0.0, units='m', desc='Hub-height')
        self.add_input('rna_mass', val=0.0, units='kg', desc='Total tower mass')
        self.add_input('rna_I', np.zeros((6,)), units='kg*m**2', desc='mass moment of inertia of rna about tower top [xx yy zz xy xz yz]')
        self.add_input('rna_cg', np.zeros((3,)), units='m', desc='xyz-location of rna cg relative to tower top')
        
        self.add_input('tower_mass', val=0.0, units='kg', desc='Total tower mass (not including monopile)')
        self.add_input('monopile_mass', val=0.0, units='kg', desc='Monopile mass')
        self.add_input('tower_center_of_mass', val=0.0, units='m', desc='z-position of center of mass of tower')
        self.add_input('tower_I_base', np.zeros((6,)), units='kg*m**2', desc='mass moment of inertia of tower about base [xx yy zz xy xz yz]')

        self.add_output('turbine_mass', val=0.0, units='kg', desc='Total mass of tower+rna')
        self.add_output('turbine_center_of_mass', val=np.zeros((3,)), units='m', desc='xyz-position of tower+rna center of mass')
        self.add_output('turbine_I_base', np.zeros((6,)), units='kg*m**2', desc='mass moment of inertia of tower about base [xx yy zz xy xz yz]')
       
        # Derivatives
        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)
        
    def compute(self, inputs, outputs):
        outputs['turbine_mass'] = inputs['rna_mass'] + inputs['tower_mass'] + inputs['monopile_mass']
        
        cg_rna   = inputs['rna_cg'] + np.array([0.0, 0.0, inputs['hub_height']])
        cg_tower = np.array([0.0, 0.0, inputs['tower_center_of_mass']])
        outputs['turbine_center_of_mass'] = (inputs['rna_mass']*cg_rna + inputs['tower_mass']*cg_tower) / outputs['turbine_mass']

        R       = cg_rna
        I_tower = assembleI(inputs['tower_I_base'])
        I_rna   = assembleI(inputs['rna_I']) + inputs['rna_mass']*(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        outputs['turbine_I_base'] = unassembleI(I_tower + I_rna)

        

        

        
class TowerPreFrame(ExplicitComponent):
    def initialize(self):
        self.options.declare('nFull')
        self.options.declare('nPoints')
        self.options.declare('monopile', default=False)
    
    def setup(self):
        nPoints = self.options['nPoints']
        nFull   = self.options['nFull']
        nRefine = int( (nFull-1)/(nPoints-1) )
        
        self.add_input('z', np.zeros(nFull), units='m', desc='location along tower. start at bottom and go to top')
        self.add_input('d', np.zeros(nFull), units='m', desc='diameter along tower')

        # extra mass
        self.add_input('mass', 0.0, units='kg', desc='added mass')
        self.add_input('mI', np.zeros((6,)), units='kg*m**2', desc='mass moment of inertia about some point p [xx yy zz xy xz yz]')
        self.add_input('mrho', np.zeros((3,)), units='m', desc='xyz-location of p relative to node')
        self.add_input('transition_piece_mass', 0.0, units='kg', desc='point mass of transition piece')
        self.add_input('gravity_foundation_mass', 0.0, units='kg', desc='point mass of transition piece')
        self.add_input('transition_piece_height', 0.0, units='m', desc='height of transition piece above water line')
        self.add_input('foundation_height', 0.0, units='m', desc='height of foundation (0.0 for land, -water_depth for fixed bottom)')
        
        # point loads
        self.add_input('rna_F', np.zeros((3,)), units='N', desc='rna force')
        self.add_input('rna_M', np.zeros((3,)), units='N*m', desc='rna moment')

        # Monopile handling
        self.add_input('k_monopile', np.zeros(6), units='N/m', desc='Stiffness BCs for ocean soil.  Only used if monoflag inputis True')
        
        # spring reaction data.  Use float('inf') for rigid constraints.
        nK = nRefine+1 if self.options['monopile'] else 1
        self.add_output('kidx', np.zeros(nK, dtype=np.int_), desc='indices of z where external stiffness reactions should be applied.')
        self.add_output('kx', np.zeros(nK), units='N/m', desc='spring stiffness in x-direction')
        self.add_output('ky', np.zeros(nK), units='N/m', desc='spring stiffness in y-direction')
        self.add_output('kz', np.zeros(nK), units='N/m', desc='spring stiffness in z-direction')
        self.add_output('ktx', np.zeros(nK), units='N/m', desc='spring stiffness in theta_x-rotation')
        self.add_output('kty', np.zeros(nK), units='N/m', desc='spring stiffness in theta_y-rotation')
        self.add_output('ktz', np.zeros(nK), units='N/m', desc='spring stiffness in theta_z-rotation')
        
        # extra mass
        nMass = 3
        self.add_output('midx', np.zeros(nMass, dtype=np.int_), desc='indices where added mass should be applied.')
        self.add_output('m', np.zeros(nMass), units='kg', desc='added mass')
        self.add_output('mIxx', np.zeros(nMass), units='kg*m**2', desc='x mass moment of inertia about some point p')
        self.add_output('mIyy', np.zeros(nMass), units='kg*m**2', desc='y mass moment of inertia about some point p')
        self.add_output('mIzz', np.zeros(nMass), units='kg*m**2', desc='z mass moment of inertia about some point p')
        self.add_output('mIxy', np.zeros(nMass), units='kg*m**2', desc='xy mass moment of inertia about some point p')
        self.add_output('mIxz', np.zeros(nMass), units='kg*m**2', desc='xz mass moment of inertia about some point p')
        self.add_output('mIyz', np.zeros(nMass), units='kg*m**2', desc='yz mass moment of inertia about some point p')
        self.add_output('mrhox', np.zeros(nMass), units='m', desc='x-location of p relative to node')
        self.add_output('mrhoy', np.zeros(nMass), units='m', desc='y-location of p relative to node')
        self.add_output('mrhoz', np.zeros(nMass), units='m', desc='z-location of p relative to node')

        # point loads (if addGravityLoadForExtraMass=True be sure not to double count by adding those force here also)
        nPL = 1
        self.add_output('plidx', np.zeros(nPL, dtype=np.int_), desc='indices where point loads should be applied.')
        self.add_output('Fx', np.zeros(nPL), units='N', desc='point force in x-direction')
        self.add_output('Fy', np.zeros(nPL), units='N', desc='point force in y-direction')
        self.add_output('Fz', np.zeros(nPL), units='N', desc='point force in z-direction')
        self.add_output('Mxx', np.zeros(nPL), units='N*m', desc='point moment about x-axis')
        self.add_output('Myy', np.zeros(nPL), units='N*m', desc='point moment about y-axis')
        self.add_output('Mzz', np.zeros(nPL), units='N*m', desc='point moment about z-axis')

        #self.declare_partials('m','mass')
        #self.declare_partials(['mIxx','mIyy','mIzz','mIxy','mIxz','mIyz'], 'mI')
        #self.declare_partials(['Fx','Fy','Fz'], 'rna_F')
        #self.declare_partials(['Mxx','Myy','Mzz'], 'rna_M')

        
    def compute(self, inputs, outputs):
        nPoints = self.options['nPoints']
        nFull   = self.options['nFull']
        nRefine = int( (nFull-1)/(nPoints-1) )
        
        # Prepare RNA, transition piece, and gravity foundation (if any applicable) for "extra node mass"
        itrans = find_nearest(inputs['z'], inputs['transition_piece_height'])
        rtrans = 0.5*inputs['d'][itrans]
        mtrans = inputs['transition_piece_mass']
        Itrans = mtrans*rtrans**2. * np.r_[0.5, 0.5, 1.0, np.zeros(3)] # shell
        rgrav  = 0.5*inputs['d'][0]
        mgrav  = inputs['gravity_foundation_mass']
        Igrav  = mgrav*rgrav**2. * np.r_[0.25, 0.25, 0.5, np.zeros(3)] # disk
        # Note, need len()-1 because Frame3DD crashes if mass add at end
        outputs['midx']  = np.array([ len(inputs['z'])-1, itrans, 0 ], dtype=np.int_)
        outputs['m']     = np.array([ inputs['mass'],  mtrans,   mgrav ]).flatten()
        outputs['mIxx']  = np.array([ inputs['mI'][0], Itrans[0], Igrav[0] ]).flatten()
        outputs['mIyy']  = np.array([ inputs['mI'][1], Itrans[1], Igrav[1] ]).flatten()
        outputs['mIzz']  = np.array([ inputs['mI'][2], Itrans[2], Igrav[2] ]).flatten()
        outputs['mIxy']  = np.array([ inputs['mI'][3], Itrans[3], Igrav[3] ]).flatten()
        outputs['mIxz']  = np.array([ inputs['mI'][4], Itrans[4], Igrav[4] ]).flatten()
        outputs['mIyz']  = np.array([ inputs['mI'][5], Itrans[5], Igrav[5] ]).flatten()
        outputs['mrhox'] = np.array([ inputs['mrho'][0], 0.0, 0.0 ]).flatten()
        outputs['mrhoy'] = np.array([ inputs['mrho'][1], 0.0, 0.0 ]).flatten()
        outputs['mrhoz'] = np.array([ inputs['mrho'][2], 0.0, 0.0 ]).flatten()

        # Prepare point forces at RNA node
        outputs['plidx'] = np.array([ len(inputs['z'])-1 ], dtype=np.int_) # -1 b/c same reason as above
        outputs['Fx']    = np.array([ inputs['rna_F'][0] ]).flatten()
        outputs['Fy']    = np.array([ inputs['rna_F'][1] ]).flatten()
        outputs['Fz']    = np.array([ inputs['rna_F'][2] ]).flatten()
        outputs['Mxx']   = np.array([ inputs['rna_M'][0] ]).flatten()
        outputs['Myy']   = np.array([ inputs['rna_M'][1] ]).flatten()
        outputs['Mzz']   = np.array([ inputs['rna_M'][2] ]).flatten()

        # Prepare for reactions: rigid at tower base
        if self.options['monopile']:
            # Based on discretization, always same number of points in pile section
            kmono  = inputs['k_monopile']
            indvec = np.arange(0, nRefine+1, dtype=np.int_)
            outputs['kidx'] = indvec
            outputs['kx']   = kmono[0]*np.ones(indvec.shape)
            outputs['ky']   = kmono[2]*np.ones(indvec.shape)
            outputs['kz']   = kmono[4]*np.ones(indvec.shape)
            outputs['ktx']  = kmono[1]*np.ones(indvec.shape)
            outputs['kty']  = kmono[3]*np.ones(indvec.shape)
            outputs['ktz']  = kmono[5]*np.ones(indvec.shape)
        else:
            outputs['kidx'] = np.array([ 0 ], dtype=np.int_)
            outputs['kx']   = np.array([ 1.e16 ])
            outputs['ky']   = np.array([ 1.e16 ])
            outputs['kz']   = np.array([ 1.e16 ])
            outputs['ktx']  = np.array([ 1.e16 ])
            outputs['kty']  = np.array([ 1.e16 ])
            outputs['ktz']  = np.array([ 1.e16 ])

    '''
    def compute_partials(self, inputs, J, discrete_inputs):
        
        J['m','mass']    = 1.0
        J['mIxx','mI']   = np.eye(6)[0,:]
        J['mIyy','mI']   = np.eye(6)[1,:]
        J['mIzz','mI']   = np.eye(6)[2,:]
        J['mIxy','mI']   = np.eye(6)[3,:]
        J['mIxz','mI']   = np.eye(6)[4,:]
        J['mIyz','mI']   = np.eye(6)[5,:]
        J['Fx','rna_F']  = np.eye(3)[0,:]
        J['Fy','rna_F']  = np.eye(3)[2,:]
        J['Fz','rna_F']  = np.eye(3)[2,:]
        J['Mxx','rna_M'] = np.eye(3)[0,:]
        J['Myy','rna_M'] = np.eye(3)[2,:]
        J['Mzz','rna_M'] = np.eye(3)[2,:]
    '''

        
class TowerPostFrame(ExplicitComponent):
    def initialize(self):
        self.options.declare('nFull')
        #self.options.declare('nDEL')

    def setup(self):
        nFull = self.options['nFull']
        #nDEL  = self.options['nDEL']

        # effective geometry -- used for handbook methods to estimate hoop stress, buckling, fatigue
        self.add_input('z', np.zeros(nFull), units='m', desc='location along tower. start at bottom and go to top')
        self.add_input('d', np.zeros(nFull), units='m', desc='effective tower diameter for section')
        self.add_input('t', np.zeros(nFull-1), units='m', desc='effective shell thickness for section')
        self.add_input('L_reinforced', 0.0, units='m', desc='buckling length')

        # Material properties
        self.add_input('E', 0.0, units='N/m**2', desc='modulus of elasticity')

        # Processed Frame3DD outputs
        self.add_input('Fz', np.zeros(nFull-1), units='N', desc='Axial foce in vertical z-direction in cylinder structure.')
        self.add_input('Mxx', np.zeros(nFull-1), units='N*m', desc='Moment about x-axis in cylinder structure.')
        self.add_input('Myy', np.zeros(nFull-1), units='N*m', desc='Moment about y-axis in cylinder structure.')
        self.add_input('axial_stress', val=np.zeros(nFull-1), units='N/m**2', desc='axial stress in tower elements')
        self.add_input('shear_stress', val=np.zeros(nFull-1), units='N/m**2', desc='shear stress in tower elements')
        self.add_input('hoop_stress' , val=np.zeros(nFull-1), units='N/m**2', desc='hoop stress in tower elements')
        self.add_input('top_deflection_in', 0.0, units='m', desc='Deflection of tower top in yaw-aligned +x direction')

        # safety factors
        self.add_input('gamma_f', 1.35, desc='safety factor on loads')
        self.add_input('gamma_m', 1.1, desc='safety factor on materials')
        self.add_input('gamma_n', 1.0, desc='safety factor on consequence of failure')
        self.add_input('gamma_b', 1.1, desc='buckling safety factor')
        self.add_input('sigma_y', 0.0, units='N/m**2', desc='yield stress')
        self.add_input('gamma_fatigue', 1.755, desc='total safety factor for fatigue')

        # fatigue parameters
        self.add_input('life', 20.0, desc='fatigue life of tower')
        self.add_input('m_SN', 4, desc='slope of S/N curve')
        self.add_input('DC', 80.0, desc='standard value of stress')
        #self.add_input('z_DEL', np.zeros(nDEL), units='m', desc='absolute z coordinates of corresponding fatigue parameters')
        #self.add_input('M_DEL', np.zeros(nDEL), desc='fatigue parameters at corresponding z coordinates')

        # Frequencies
        self.add_input('f1', 0.0, units='Hz', desc='First natural frequency')
        self.add_input('f2', 0.0, units='Hz', desc='Second natural frequency')
        
        # outputs
        self.add_output('structural_frequencies', np.zeros(2), units='Hz', desc='First and second natural frequency')
        self.add_output('top_deflection', 0.0, units='m', desc='Deflection of tower top in yaw-aligned +x direction')
        self.add_output('stress', np.zeros(nFull-1), desc='Von Mises stress utilization along tower at specified locations.  incudes safety factor.')
        self.add_output('shell_buckling', np.zeros(nFull-1), desc='Shell buckling constraint.  Should be < 1 for feasibility.  Includes safety factors')
        self.add_output('global_buckling', np.zeros(nFull-1), desc='Global buckling constraint.  Should be < 1 for feasibility.  Includes safety factors')
        #self.add_output('damage', np.zeros(nFull-1), desc='Fatigue damage at each tower section')
        self.add_output('turbine_F', val=np.zeros(3), units='N', desc='Total force on tower+rna')
        self.add_output('turbine_M', val=np.zeros(3), units='N*m', desc='Total x-moment on tower+rna measured at base')
        
        # Derivatives
        # self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs):
        # Unpack some variables
        axial_stress = inputs['axial_stress']
        shear_stress = inputs['shear_stress']
        hoop_stress  = inputs['hoop_stress']
        sigma_y      = inputs['sigma_y'] * np.ones(axial_stress.shape)
        E            = inputs['E'] * np.ones(axial_stress.shape)
        L_reinforced = inputs['L_reinforced'] * np.ones(axial_stress.shape)
        d,_          = nodal2sectional(inputs['d'])
        z_section,_  = nodal2sectional(inputs['z'])

        # Frequencies
        outputs['structural_frequencies'] = np.zeros(2)
        outputs['structural_frequencies'][0] = inputs['f1']
        outputs['structural_frequencies'][1] = inputs['f2']

        # Deflection
        outputs['top_deflection'] = inputs['top_deflection_in']
        
        # von mises stress
        outputs['stress'] = Util.vonMisesStressUtilization(axial_stress, hoop_stress, shear_stress,
                      inputs['gamma_f']*inputs['gamma_m']*inputs['gamma_n'], sigma_y)

        # shell buckling
        outputs['shell_buckling'] = Util.shellBucklingEurocode(d, inputs['t'], axial_stress, hoop_stress,
                                                                shear_stress, L_reinforced, E, sigma_y, inputs['gamma_f'], inputs['gamma_b'])

        # global buckling
        tower_height = inputs['z'][-1] - inputs['z'][0]
        M = np.sqrt(inputs['Mxx']**2 + inputs['Myy']**2)
        outputs['global_buckling'] = Util.bucklingGL(d, inputs['t'], inputs['Fz'], M, tower_height, E,
                                                      sigma_y, inputs['gamma_f'], inputs['gamma_b'])

        # fatigue
        N_DEL = 365.0*24.0*3600.0*inputs['life'] * np.ones(len(inputs['t']))
        #outputs['damage'] = np.zeros(N_DEL.shape)

        #if any(inputs['M_DEL']):
        #    M_DEL = np.interp(z_section, inputs['z_DEL'], inputs['M_DEL'])

        #    outputs['damage'] = Util.fatigue(M_DEL, N_DEL, d, inputs['t'], inputs['m_SN'],
        #                                      inputs['DC'], inputs['gamma_fatigue'], stress_factor=1.0, weld_factor=True)

# -----------------
#  Assembly
# -----------------

class TowerLeanSE(Group):

    def initialize(self):
        self.options.declare('nPoints')
        self.options.declare('nFull')
        self.options.declare('topLevelFlag', default=True)
        self.options.declare('monopile', default=False)
        
    def setup(self):
        nPoints      = self.options['nPoints']
        nFull        = self.options['nFull']
        topLevelFlag = self.options['topLevelFlag']
        nRefine      = int( (nFull-1)/(nPoints-1) )
        monopile     = self.options['monopile']
        
        # Independent variables that are unique to TowerSE
        towerIndeps = IndepVarComp()
        towerIndeps.add_output('tower_outer_diameter', np.zeros(nPoints), units='m')
        towerIndeps.add_output('tower_section_height', np.zeros(nPoints-1), units='m')
        towerIndeps.add_output('tower_wall_thickness', np.zeros(nPoints-1), units='m')
        towerIndeps.add_output('tower_buckling_length', 0.0, units='m')
        towerIndeps.add_output('tower_outfitting_factor', 0.0)
        towerIndeps.add_output('gravity_foundation_mass', 0.0, units='kg')
        towerIndeps.add_output('transition_piece_mass', 0.0, units='kg')
        towerIndeps.add_output('transition_piece_height', 0.0, units='m')
        towerIndeps.add_output('suctionpile_depth', 0.0, units='m')
        towerIndeps.add_output('suctionpile_depth_diam_ratio', 0.0)
        self.add_subsystem('towerIndepsLean', towerIndeps, promotes=['*'])

        # Independent variables that may be duplicated at higher levels of aggregation
        if topLevelFlag:
            sharedIndeps = IndepVarComp()
            sharedIndeps.add_output('hub_height', 0.0, units='m')
            sharedIndeps.add_output('material_density', 0.0, units='kg/m**3')
            self.add_subsystem('sharedIndepsLean', sharedIndeps, promotes=['*'])
        
        # All the static components
        self.add_subsystem('predisc', MonopileFoundation(monopile=monopile, nPoints=nPoints), promotes=['*'])
        # If doing fixed bottom monopile, we add an additional point for the pile (even for gravity foundations)
        if monopile:
            nPoints += 1
            nFull   += nRefine
            
        # Promote all but foundation_height so that we can override
        self.add_subsystem('geometry', CylinderDiscretization(nPoints=nPoints, nRefine=nRefine), promotes=['section_height','wall_thickness','diameter',
                                                                                                           'z_param','z_full','d_full','t_full'])
        
        self.add_subsystem('tgeometry', TowerDiscretization(nPoints=nPoints), promotes=['*'])
        
        self.add_subsystem('cm', CylinderMass(nPoints=nFull), promotes=['material_density','z_full','d_full','t_full',
                                                                        'material_cost_rate','labor_cost_rate','painting_cost_rate'])
        self.add_subsystem('tm', TowerMass(nFull=nFull), promotes=['z_full',
                                                                   'tower_mass','tower_center_of_mass','tower_section_center_of_mass','tower_I_base',
                                                                   'tower_raw_cost','gravity_foundation_mass','foundation_height',
                                                                   'transition_piece_mass','transition_piece_height',
                                                                   'monopile_mass','monopile_cost','monopile_length'])
        self.add_subsystem('gc', Util.GeometricConstraints(nPoints=nPoints), promotes=['min_d_to_t','max_taper','manufacturability','weldability','slope'])
        self.add_subsystem('turb', TurbineMass(), promotes=['turbine_mass','monopile_mass',
                                                            'tower_mass','tower_center_of_mass','tower_I_base',
                                                            'rna_mass', 'rna_cg', 'rna_I','hub_height'])
        
        # Connections for geometry and mass
        self.connect('foundation_height_out', 'geometry.foundation_height')
        self.connect('section_height_out', 'section_height')
        self.connect('outer_diameter_out', ['diameter', 'gc.d'])
        self.connect('wall_thickness_out', ['wall_thickness', 'gc.t'])
        self.connect('tower_outfitting_factor', 'cm.outfitting_factor')

        self.connect('cm.mass', 'tm.cylinder_mass')
        self.connect('cm.cost', 'tm.cylinder_cost')
        self.connect('cm.center_of_mass', 'tm.cylinder_center_of_mass')
        self.connect('cm.section_center_of_mass','tm.cylinder_section_center_of_mass')
        self.connect('cm.I_base','tm.cylinder_I_base')

        
class TowerSE(Group):

    def initialize(self):
        self.options.declare('nLC')
        self.options.declare('nPoints')
        self.options.declare('nFull')
        #self.options.declare('nDEL')
        self.options.declare('wind', default='')
        self.options.declare('topLevelFlag', default=True)
        self.options.declare('monopile', default=False)
    
    def setup(self):
        nLC           = self.options['nLC']
        nPoints       = self.options['nPoints']
        nFull         = self.options['nFull']
        # nDEL          = self.options['nDEL']
        wind          = self.options['wind']
        topLevelFlag  = self.options['topLevelFlag']
        monopile      = self.options['monopile']
        nRefine       = int( (nFull-1)/(nPoints-1) )
        nK = nRefine+1 if self.options['monopile'] else 1
        
        # Independent variables that are unique to TowerSE
        towerIndeps = IndepVarComp()
        towerIndeps.add_output('foundation_height', 0.0, units='m')
        #towerIndeps.add_output('tower_M_DEL', np.zeros(nDEL))
        #towerIndeps.add_output('tower_z_DEL', np.zeros(nDEL), units='m')
        towerIndeps.add_output('tower_force_discretization', 5.0)
        towerIndeps.add_output('soil_G', 0.0, units='N/m**2')
        towerIndeps.add_output('soil_nu', 0.0)
        towerIndeps.add_discrete_output('tower_add_gravity', True)
        towerIndeps.add_discrete_output('shear', True)
        towerIndeps.add_discrete_output('geom', True)
        towerIndeps.add_discrete_output('nM', 2)
        towerIndeps.add_discrete_output('Mmethod', 1)
        towerIndeps.add_discrete_output('lump', 0)
        towerIndeps.add_output('tol', 1e-9)
        towerIndeps.add_output('shift', 0.0)
        towerIndeps.add_output('DC', 0.0)
        towerIndeps.add_output('water_density', 1025.0, units='kg/m**3')
        towerIndeps.add_output('water_viscosity', 8.9e-4, units='kg/m/s')
        towerIndeps.add_output('significant_wave_height', 0.0, units='m')
        towerIndeps.add_output('significant_wave_period', 0.0, units='s')
        towerIndeps.add_output('wave_beta', 0.0, units='deg')
        self.add_subsystem('towerIndeps', towerIndeps, promotes=['*'])

        # Independent variables that may be duplicated at higher levels of aggregation
        if topLevelFlag:
            sharedIndeps = IndepVarComp()
            sharedIndeps.add_output('air_density', 1.225, units='kg/m**3')
            sharedIndeps.add_output('air_viscosity', 1.81206e-5, units='kg/m/s')
            sharedIndeps.add_output('shearExp', 0.0)
            sharedIndeps.add_output('wind_reference_height', 0.0, units='m')
            sharedIndeps.add_output('wind_z0', 0.0, units='m')
            sharedIndeps.add_output('wind_beta', 0.0, units='deg')
            sharedIndeps.add_output('cd_usr', -1.)
            sharedIndeps.add_output('yaw', 0.0, units='deg')
            sharedIndeps.add_output('E', 0.0, units='N/m**2')
            sharedIndeps.add_output('G', 0.0, units='N/m**2')
            sharedIndeps.add_output('sigma_y', 0.0, units='N/m**2')
            sharedIndeps.add_output('rna_mass', 0.0, units='kg')
            sharedIndeps.add_output('rna_cg', np.zeros(3), units='m')
            sharedIndeps.add_output('rna_I', np.zeros(6), units='kg*m**2')
            sharedIndeps.add_output('gamma_f', 0.0)
            sharedIndeps.add_output('gamma_m', 0.0)
            sharedIndeps.add_output('gamma_n', 0.0)
            sharedIndeps.add_output('gamma_b', 0.0)
            sharedIndeps.add_output('gamma_fatigue', 0.0)
            sharedIndeps.add_output('life', 0.0)
            sharedIndeps.add_output('m_SN', 0.0)
            self.add_subsystem('sharedIndeps', sharedIndeps, promotes=['*'])

        self.add_subsystem('geom', TowerLeanSE(nPoints=nPoints, nFull=nFull, topLevelFlag=topLevelFlag, monopile=monopile), promotes=['*'])
        # If doing fixed bottom monopile, we add an additional point for the pile (even for gravity foundations)
        if monopile:
            nPoints += 1
            nFull   += nRefine
        self.add_subsystem('props', CylindricalShellProperties(nFull=nFull))
        self.add_subsystem('soil', TowerSoil())

        # Connections for geometry and mass
        self.connect('d_full', 'props.d')
        self.connect('t_full', 'props.t')
        self.connect('d_full', 'soil.d0', src_indices=[0])
        self.connect('suctionpile_depth', 'soil.depth')
        self.connect('soil_G', 'soil.G')
        self.connect('soil_nu', 'soil.nu')
        
        # Add in all Components that drive load cases
        # Note multiple load cases have to be handled by replicating components and not groups/assemblies.
        # Replicating Groups replicates the IndepVarComps which doesn't play nicely in OpenMDAO
        for iLC in range(nLC):
            lc = '' if nLC==1 else str(iLC+1)
            
            if wind is None or wind.lower() in ['power', 'powerwind', '']:
                self.add_subsystem('wind'+lc, PowerWind(nPoints=nFull))
            elif wind.lower() == 'logwind':
                self.add_subsystem('wind'+lc, LogWind(nPoints=nFull))
            else:
                raise ValueError('Unknown wind type, '+wind)

            self.add_subsystem('windLoads'+lc, CylinderWindDrag(nPoints=nFull), promotes=['cd_usr'])

            if monopile:
                self.add_subsystem('wave'+lc, LinearWaves(nPoints=nFull), promotes=['z_floor'])
                self.add_subsystem('waveLoads'+lc, CylinderWaveDrag(nPoints=nFull), promotes=['cm','cd_usr'])

            self.add_subsystem('distLoads'+lc, AeroHydroLoads(nPoints=nFull), promotes=['yaw'])

            self.add_subsystem('pre'+lc, TowerPreFrame(nFull=nFull, nPoints=nPoints, monopile=monopile), promotes=['transition_piece_mass',
                                                                                                                   'transition_piece_height',
                                                                                                                   'gravity_foundation_mass'])
            self.add_subsystem('tower'+lc, CylinderFrame3DD(npts=nFull, nK=nK, nMass=3, nPL=1), promotes=['E','G','tol','Mmethod','geom','lump','shear',
                                                                                                         'nM','shift','sigma_y'])
            self.add_subsystem('post'+lc, TowerPostFrame(nFull=nFull), promotes=['E','sigma_y','DC','life','m_SN',
                                                                                 'gamma_b','gamma_f','gamma_fatigue','gamma_m','gamma_n'])
            
            self.connect('z_full', ['wind'+lc+'.z', 'windLoads'+lc+'.z', 'distLoads'+lc+'.z', 'pre'+lc+'.z', 'tower'+lc+'.z', 'post'+lc+'.z'])
            self.connect('d_full', ['windLoads'+lc+'.d', 'tower'+lc+'.d', 'pre'+lc+'.d', 'post'+lc+'.d'])
            if monopile:
                self.connect('z_full', ['wave'+lc+'.z', 'waveLoads'+lc+'.z'])
                self.connect('d_full', 'waveLoads'+lc+'.d')

            if topLevelFlag:
                self.connect('rna_mass', 'pre'+lc+'.mass')
                self.connect('rna_cg', 'pre'+lc+'.mrho')
                self.connect('rna_I', 'pre'+lc+'.mI')
                self.connect('material_density', 'tower'+lc+'.rho')

            self.connect('pre'+lc+'.kidx', 'tower'+lc+'.kidx')
            self.connect('pre'+lc+'.kx', 'tower'+lc+'.kx')
            self.connect('pre'+lc+'.ky', 'tower'+lc+'.ky')
            self.connect('pre'+lc+'.kz', 'tower'+lc+'.kz')
            self.connect('pre'+lc+'.ktx', 'tower'+lc+'.ktx')
            self.connect('pre'+lc+'.kty', 'tower'+lc+'.kty')
            self.connect('pre'+lc+'.ktz', 'tower'+lc+'.ktz')
            self.connect('pre'+lc+'.midx', 'tower'+lc+'.midx')
            self.connect('pre'+lc+'.m', 'tower'+lc+'.m')
            self.connect('pre'+lc+'.mIxx', 'tower'+lc+'.mIxx')
            self.connect('pre'+lc+'.mIyy', 'tower'+lc+'.mIyy')
            self.connect('pre'+lc+'.mIzz', 'tower'+lc+'.mIzz')
            self.connect('pre'+lc+'.mIxy', 'tower'+lc+'.mIxy')
            self.connect('pre'+lc+'.mIxz', 'tower'+lc+'.mIxz')
            self.connect('pre'+lc+'.mIyz', 'tower'+lc+'.mIyz')
            self.connect('pre'+lc+'.mrhox', 'tower'+lc+'.mrhox')
            self.connect('pre'+lc+'.mrhoy', 'tower'+lc+'.mrhoy')
            self.connect('pre'+lc+'.mrhoz', 'tower'+lc+'.mrhoz')

            self.connect('pre'+lc+'.plidx', 'tower'+lc+'.plidx')
            self.connect('pre'+lc+'.Fx', 'tower'+lc+'.Fx')
            self.connect('pre'+lc+'.Fy', 'tower'+lc+'.Fy')
            self.connect('pre'+lc+'.Fz', 'tower'+lc+'.Fz')
            self.connect('pre'+lc+'.Mxx', 'tower'+lc+'.Mxx')
            self.connect('pre'+lc+'.Myy', 'tower'+lc+'.Myy')
            self.connect('pre'+lc+'.Mzz', 'tower'+lc+'.Mzz')
            self.connect('tower_force_discretization', 'tower'+lc+'.dx')
            self.connect('tower_add_gravity', 'tower'+lc+'.addGravityLoadForExtraMass')
            self.connect('t_full', ['tower'+lc+'.t','post'+lc+'.t'])
            self.connect('soil.k', 'pre'+lc+'.k_monopile')

            self.connect('tower'+lc+'.f1', 'post'+lc+'.f1')
            self.connect('tower'+lc+'.f2', 'post'+lc+'.f2')
            self.connect('tower'+lc+'.Fz_out', 'post'+lc+'.Fz')
            self.connect('tower'+lc+'.Mxx_out', 'post'+lc+'.Mxx')
            self.connect('tower'+lc+'.Myy_out', 'post'+lc+'.Myy')
            self.connect('tower'+lc+'.axial_stress', 'post'+lc+'.axial_stress')
            self.connect('tower'+lc+'.shear_stress', 'post'+lc+'.shear_stress')
            self.connect('tower'+lc+'.hoop_stress_euro', 'post'+lc+'.hoop_stress')
            self.connect('tower'+lc+'.top_deflection', 'post'+lc+'.top_deflection_in')
        
            # connections to wind, wave
            self.connect('wind'+lc+'.U', 'windLoads'+lc+'.U')
            if topLevelFlag:
                self.connect('wind_reference_height', 'wind'+lc+'.zref')
                self.connect('wind_z0', 'wind'+lc+'.z0')
                if monopile:
                    self.connect('wind_z0', 'wave'+lc+'.z_surface')
                #self.connect('z_floor', 'waveLoads'+lc+'.wlevel')
                if wind=='PowerWind':
                    self.connect('shearExp', 'wind'+lc+'.shearExp')
                
                # connections to windLoads1
                self.connect('air_density', 'windLoads'+lc+'.rho')
                self.connect('air_viscosity', 'windLoads'+lc+'.mu')
                self.connect('wind_beta', 'windLoads'+lc+'.beta')

            if monopile:
                # connections to waveLoads1
                self.connect('water_density', ['wave'+lc+'.rho', 'waveLoads'+lc+'.rho'])
                self.connect('water_viscosity', 'waveLoads'+lc+'.mu')
                self.connect('wave_beta', 'waveLoads'+lc+'.beta')
                self.connect('significant_wave_height', 'wave'+lc+'.hmax')
                self.connect('significant_wave_period', 'wave'+lc+'.T')
                self.connect('foundation_height', 'z_floor')
                    
                self.connect('wave'+lc+'.U', 'waveLoads'+lc+'.U')
                self.connect('wave'+lc+'.A', 'waveLoads'+lc+'.A')
                self.connect('wave'+lc+'.p', 'waveLoads'+lc+'.p')

            # connections to distLoads1
            self.connect('windLoads'+lc+'.windLoads_Px', 'distLoads'+lc+'.windLoads_Px')
            self.connect('windLoads'+lc+'.windLoads_Py', 'distLoads'+lc+'.windLoads_Py')
            self.connect('windLoads'+lc+'.windLoads_Pz', 'distLoads'+lc+'.windLoads_Pz')
            self.connect('windLoads'+lc+'.windLoads_qdyn', 'distLoads'+lc+'.windLoads_qdyn')
            self.connect('windLoads'+lc+'.windLoads_beta', 'distLoads'+lc+'.windLoads_beta')
            #self.connect('windLoads'+lc+'.windLoads_Px0', 'distLoads'+lc+'.windLoads_Px0')
            #self.connect('windLoads'+lc+'.windLoads_Py0', 'distLoads'+lc+'.windLoads_Py0')
            #self.connect('windLoads'+lc+'.windLoads_Pz0', 'distLoads'+lc+'.windLoads_Pz0')
            #self.connect('windLoads'+lc+'.windLoads_qdyn0', 'distLoads'+lc+'.windLoads_qdyn0')
            #self.connect('windLoads'+lc+'.windLoads_beta0', 'distLoads'+lc+'.windLoads_beta0')
            self.connect('windLoads'+lc+'.windLoads_z', 'distLoads'+lc+'.windLoads_z')
            self.connect('windLoads'+lc+'.windLoads_d', 'distLoads'+lc+'.windLoads_d')

            if monopile:
                self.connect('waveLoads'+lc+'.waveLoads_Px', 'distLoads'+lc+'.waveLoads_Px')
                self.connect('waveLoads'+lc+'.waveLoads_Py', 'distLoads'+lc+'.waveLoads_Py')
                self.connect('waveLoads'+lc+'.waveLoads_Pz', 'distLoads'+lc+'.waveLoads_Pz')
                self.connect('waveLoads'+lc+'.waveLoads_pt', 'distLoads'+lc+'.waveLoads_qdyn')
                self.connect('waveLoads'+lc+'.waveLoads_beta', 'distLoads'+lc+'.waveLoads_beta')
                #self.connect('waveLoads'+lc+'.waveLoads_Px0', 'distLoads'+lc+'.waveLoads_Px0')
                #self.connect('waveLoads'+lc+'.waveLoads_Py0', 'distLoads'+lc+'.waveLoads_Py0')
                #self.connect('waveLoads'+lc+'.waveLoads_Pz0', 'distLoads'+lc+'.waveLoads_Pz0')
                #self.connect('waveLoads'+lc+'.waveLoads_qdyn0', 'distLoads'+lc+'.waveLoads_qdyn0')
                #self.connect('waveLoads'+lc+'.waveLoads_beta0', 'distLoads'+lc+'.waveLoads_beta0')
                self.connect('waveLoads'+lc+'.waveLoads_z', 'distLoads'+lc+'.waveLoads_z')
                self.connect('waveLoads'+lc+'.waveLoads_d', 'distLoads'+lc+'.waveLoads_d')

            # Tower connections
            self.connect('tower_buckling_length', ['tower'+lc+'.L_reinforced', 'post'+lc+'.L_reinforced'])
            #self.connect('tower_M_DEL', 'post'+lc+'.M_DEL')
            #self.connect('tower_z_DEL', 'post'+lc+'.z_DEL')

            self.connect('props.Az', 'tower'+lc+'.Az')
            self.connect('props.Asx', 'tower'+lc+'.Asx')
            self.connect('props.Asy', 'tower'+lc+'.Asy')
            self.connect('props.Jz', 'tower'+lc+'.Jz')
            self.connect('props.Ixx', 'tower'+lc+'.Ixx')
            self.connect('props.Iyy', 'tower'+lc+'.Iyy')

            self.connect('distLoads'+lc+'.Px',   'tower'+lc+'.Px')
            self.connect('distLoads'+lc+'.Py',   'tower'+lc+'.Py')
            self.connect('distLoads'+lc+'.Pz',   'tower'+lc+'.Pz')
            self.connect('distLoads'+lc+'.qdyn', 'tower'+lc+'.qdyn')

        
if __name__ == '__main__':
    # --- tower setup ------
    from wisdem.commonse.environment import PowerWind
    from wisdem.commonse.environment import LogWind

    optimize_flag = False

    # --- geometry ----
    h_param = np.diff(np.array([0.0, 43.8, 87.6]))
    d_param = np.array([6.0, 4.935, 3.87])
    t_param = 1.3*np.array([0.025, 0.021])
    z_foundation = 0.0
    L_reinforced = 30.0  # [m] buckling length
    theta_stress = 0.0
    yaw = 0.0
    Koutfitting = 1.07

    # --- material props ---
    E = 210e9
    G = 80.8e9
    rho = 8500.0
    sigma_y = 450.0e6

    # --- extra mass ----
    m = np.array([285598.8])
    mIxx = 1.14930678e+08
    mIyy = 2.20354030e+07
    mIzz = 1.87597425e+07
    mIxy = 0.0
    mIxz = 5.03710467e+05
    mIyz = 0.0
    mI = np.array([mIxx, mIyy, mIzz, mIxy, mIxz, mIyz])
    mrho = np.array([-1.13197635, 0.0, 0.50875268])
    # -----------

    # --- wind ---
    wind_zref = 90.0
    wind_z0 = 0.0
    shearExp = 0.2
    cd_usr = -1.
    # ---------------

    # --- wave ---
    hmax = 0.0
    T = 1.0
    cm = 1.0
    monopile = False
    suction_depth = 0.0
    soilG = 140e6
    soilnu = 0.4
    # ---------------

    
    # two load cases.  TODO: use a case iterator
    
    # # --- loading case 1: max Thrust ---
    wind_Uref1 = 11.73732
    Fx1 = 1284744.19620519
    Fy1 = 0.
    Fz1 = -2914124.84400512 + m*gravity
    Mxx1 = 3963732.76208099
    Myy1 = -2275104.79420872
    Mzz1 = -346781.68192839
    # # ---------------

    # # --- loading case 2: max wind speed ---
    wind_Uref2 = 70.0
    Fx2 = 930198.60063279
    Fy2 = 0.
    Fz2 = -2883106.12368949 + m*gravity
    Mxx2 = -1683669.22411597
    Myy2 = -2522475.34625363
    Mzz2 = 147301.97023764
    # # ---------------

    # --- safety factors ---
    gamma_f = 1.35
    gamma_m = 1.3
    gamma_n = 1.0
    gamma_b = 1.1
    # ---------------

    # --- fatigue ---
    z_DEL = np.array([0.000, 1.327, 3.982, 6.636, 9.291, 11.945, 14.600, 17.255, 19.909, 22.564, 25.218, 27.873, 30.527, 33.182, 35.836, 38.491, 41.145, 43.800, 46.455, 49.109, 51.764, 54.418, 57.073, 59.727, 62.382, 65.036, 67.691, 70.345, 73.000, 75.655, 78.309, 80.964, 83.618, 86.273, 87.600])
    M_DEL = 1e3*np.array([8.2940E+003, 8.1518E+003, 7.8831E+003, 7.6099E+003, 7.3359E+003, 7.0577E+003, 6.7821E+003, 6.5119E+003, 6.2391E+003, 5.9707E+003, 5.7070E+003, 5.4500E+003, 5.2015E+003, 4.9588E+003, 4.7202E+003, 4.4884E+003, 4.2577E+003, 4.0246E+003, 3.7942E+003, 3.5664E+003, 3.3406E+003, 3.1184E+003, 2.8977E+003, 2.6811E+003, 2.4719E+003, 2.2663E+003, 2.0673E+003, 1.8769E+003, 1.7017E+003, 1.5479E+003, 1.4207E+003, 1.3304E+003, 1.2780E+003, 1.2673E+003, 1.2761E+003])
    nDEL = len(z_DEL)
    gamma_fatigue = 1.35*1.3*1.0
    life = 20.0
    m_SN = 4
    # ---------------


    # --- constraints ---
    min_d_to_t   = 120.0
    max_taper    = 0.2
    # ---------------

    # # V_max = 80.0  # tip speed
    # # D = 126.0
    # # .freq1p = V_max / (D/2) / (2*pi)  # convert to Hz

    nPoints = len(d_param)
    nFull   = 3*(nPoints-1) + 1
    wind = 'PowerWind'
    nLC = 2
    
    prob = Problem()
    prob.model = TowerSE(nLC=nLC, nPoints=nPoints, nFull=nFull, wind=wind, topLevelFlag=True, monopile=monopile)

    if optimize_flag:

        # --- optimizer imports ---
        from openmdao.api import ScipyOptimizeDriver, SqliteRecorder
        # ----------------------

        # --- Setup Pptimizer ---
        prob.driver = ScipyOptimizeDriver()
        prob.driver.options['optimizer'] = 'SLSQP'
        # ----------------------

        # --- Objective ---
        prob.model.add_objective('tower_mass', scaler=1e-6)
        # ----------------------

        # --- Design Variables ---
        prob.model.add_design_var('tower_outer_diameter', lower=3.87, upper=6.0, indices=[m for m in range(nPoints-1)])
        prob.model.add_design_var('tower_wall_thickness', lower=4e-3, upper=2e-1)
        prob.model.add_design_var('tower_section_height', lower=2.0, upper=80.0)
        # ----------------------

        # --- recorder ---
        frecord_opt = 'tower_example-optimizer.sql'
        frecord = 'tower_example.sql'
        prob.driver.add_recorder(SqliteRecorder(frecord_opt))
        prob.add_recorder(SqliteRecorder(frecord))
        # ----------------------

        # --- Constraints ---
        prob.model.add_constraint('height_constraint',    lower=-1e-2,upper=1.e-2)
        prob.model.add_constraint('post1.stress',          upper=1.0)
        prob.model.add_constraint('post1.global_buckling', upper=1.0)
        prob.model.add_constraint('post1.shell_buckling',  upper=1.0)
        prob.model.add_constraint('post2.stress',          upper=1.0)
        prob.model.add_constraint('post2.global_buckling', upper=1.0)
        prob.model.add_constraint('post2.shell_buckling',  upper=1.0)
        prob.model.add_constraint('weldability',          upper=0.0)
        prob.model.add_constraint('manufacturability',    lower=0.0)
        prob.model.add_constraint('slope',                upper=1.0)
        prob.model.add_constraint('tower1.f1',             lower=0.3)
        # ----------------------

    prob.setup()

    if wind=='PowerWind':
        prob['shearExp'] = shearExp

    # assign values to params

    # --- geometry ----
    prob['hub_height'] = h_param.sum()
    prob['foundation_height'] = 0.0
    prob['tower_section_height'] = h_param
    prob['tower_outer_diameter'] = d_param
    prob['tower_wall_thickness'] = t_param
    prob['tower_buckling_length'] = L_reinforced
    prob['tower_outfitting_factor'] = Koutfitting
    prob['yaw'] = yaw
    prob['suctionpile_depth'] = suction_depth
    prob['suctionpile_depth_diam_ratio'] = 3.35
    prob['soil_G'] = soilG
    prob['soil_nu'] = soilnu
    # --- material props ---
    prob['E'] = E
    prob['G'] = G
    prob['material_density'] = rho
    prob['sigma_y'] = sigma_y

    # --- extra mass ----
    prob['rna_mass'] = m
    prob['rna_I'] = mI
    prob['rna_cg'] = mrho
    # -----------

    # --- wind & wave ---
    prob['wind_reference_height'] = wind_zref
    prob['wind_z0'] = wind_z0
    prob['cd_usr'] = cd_usr
    prob['air_density'] = 1.225
    prob['air_viscosity'] = 1.7934e-5
    prob['water_density'] = 1025.0
    prob['water_viscosity'] = 1.3351e-3
    prob['wind_beta'] = prob['wave_beta'] = 0.0
    prob['significant_wave_height'] = hmax
    prob['significant_wave_period'] = T
    #prob['waveLoads1.U0'] = prob['waveLoads1.A0'] = prob['waveLoads1.beta0'] = prob['waveLoads2.U0'] = prob['waveLoads2.A0'] = prob['waveLoads2.beta0'] = 0.0
    # ---------------

    # --- safety factors ---
    prob['gamma_f'] = gamma_f
    prob['gamma_m'] = gamma_m
    prob['gamma_n'] = gamma_n
    prob['gamma_b'] = gamma_b
    prob['gamma_fatigue'] = gamma_fatigue
    # ---------------

    prob['DC'] = 80.0
    prob['shear'] = True
    prob['geom'] = True
    prob['tower_force_discretization'] = 5.0
    prob['nM'] = 2
    prob['Mmethod'] = 1
    prob['lump'] = 0
    prob['tol'] = 1e-9
    prob['shift'] = 0.0

    
    # --- fatigue ---
    #prob['tower_z_DEL'] = z_DEL
    #prob['tower_M_DEL'] = M_DEL
    prob['life'] = life
    prob['m_SN'] = m_SN
    # ---------------

    # --- constraints ---
    prob['min_d_to_t'] = min_d_to_t
    prob['max_taper'] = max_taper
    # ---------------


    # # --- loading case 1: max Thrust ---
    prob['wind1.Uref'] = wind_Uref1

    prob['pre1.rna_F'] = np.array([Fx1, Fy1, Fz1])
    prob['pre1.rna_M'] = np.array([Mxx1, Myy1, Mzz1])
    # # ---------------


    # # --- loading case 2: max Wind Speed ---
    prob['wind2.Uref'] = wind_Uref2

    prob['pre2.rna_F'] = np.array([Fx2, Fy2, Fz2])
    prob['pre2.rna_M' ] = np.array([Mxx2, Myy2, Mzz2])

    # # --- run ---
    if optimize_flag:
        prob.model.approx_totals()
        prob.run_driver()
    else:
        prob.run_model()

    z,_ = nodal2sectional(prob['z_full'])

    print('zs=', prob['z_full'])
    print('ds=', prob['d_full'])
    print('ts=', prob['t_full'])
    print('mass (kg) =', prob['tower_mass'])
    print('cg (m) =', prob['tower_center_of_mass'])
    print('weldability =', prob['weldability'])
    print('manufacturability =', prob['manufacturability'])
    print('\nwind: ', prob['wind1.Uref'])
    print('f1 (Hz) =', prob['tower1.f1'])
    print('top_deflection1 (m) =', prob['post1.top_deflection'])
    print('stress1 =', prob['post1.stress'])
    print('GL buckling =', prob['post1.global_buckling'])
    print('Shell buckling =', prob['post1.shell_buckling'])
    #print('damage =', prob['post1.damage'])
    print('\nwind: ', prob['wind2.Uref'])
    print('f1 (Hz) =', prob['tower2.f1'])
    print('top_deflection2 (m) =', prob['post2.top_deflection'])
    print('stress2 =', prob['post2.stress'])
    print('GL buckling =', prob['post2.global_buckling'])
    print('Shell buckling =', prob['post2.shell_buckling'])
    #print('damage =', prob['post2.damage'])


    stress1 = np.copy( prob['post1.stress'] )
    shellBuckle1 = np.copy( prob['post1.shell_buckling'] )
    globalBuckle1 = np.copy( prob['post1.global_buckling'] )
    #damage1 = np.copy( prob['post1.damage'] )

    stress2 = prob['post2.stress']
    shellBuckle2 = prob['post2.shell_buckling']
    globalBuckle2 = prob['post2.global_buckling']
    #damage2 = prob['post2.damage']

    
    import matplotlib.pyplot as plt
    plt.figure(figsize=(5.0, 3.5))
    plt.subplot2grid((3, 3), (0, 0), colspan=2, rowspan=3)
    plt.plot(stress1, z, label='stress 1')
    plt.plot(stress2, z, label='stress 2')
    plt.plot(shellBuckle1, z, label='shell buckling 1')
    plt.plot(shellBuckle2, z, label='shell buckling 2')
    plt.plot(globalBuckle1, z, label='global buckling 1')
    plt.plot(globalBuckle2, z, label='global buckling 2')
    #plt.plot(damage1, z, label='damage 1')
    #plt.plot(damage2, z, label='damage 2')
    plt.legend(bbox_to_anchor=(1.05, 1.0), loc=2)
    plt.xlabel('utilization')
    plt.ylabel('height along tower (m)')

    #plt.figure(2)
    #plt.plot(prob['d_full']/2.+max(prob['d_full']), z, 'ok')
    #plt.plot(prob['d_full']/-2.+max(prob['d_full']), z, 'ok')

    #fig = plt.figure(3)
    #ax1 = fig.add_subplot(121)
    #ax2 = fig.add_subplot(122)

    #ax1.plot(prob['wind1.U'], z)
    #ax2.plot(prob['wind2.U'], z)
    #plt.tight_layout()
    plt.show()

    print(prob['tower1.base_F'])
    print(prob['tower1.base_M'])
    print(prob['tower2.base_F'])
    print(prob['tower2.base_M'])
    # ------------




    
    



"""
===============
Rain simulation
===============

Simulates rain drops on a surface by animating the scale and opacity
of 50 scatter points.

Author: Nicolas P. Rougier
"""
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

velocities=np.load('velocity.npy')

# Create new Figure and an Axes which fills it.
fig = plt.figure(figsize=(7, 7))
ax = fig.add_subplot(111)
x = np.arange(0,29)
ax.set_xlim(0, 29), ax.set_xticks([])
ax.set_ylim(0, 29), ax.set_yticks([])

div = make_axes_locatable(ax)
cax = div.append_axes('right', '5%', '5%')

cont = ax.contourf(x, x, velocities[0, 0, :, :])
n_frames = len(velocities[:,0,0,0])
#quit()
MULT= int(1/0.05)
def update(frame_number):
    # Get an index which we can use to re-spawn the oldest raindrop.
    current_index = (frame_number * MULT) % n_frames
    cont = ax.contourf(x, x, velocities[current_index, 0, :, :], cmap=plt.cm.coolwarm)
    cax.cla()
    colorbar = fig.colorbar(cont, cax=cax)
    ax.set_title('Time is %f seconds'%(current_index*.05))
    # Update the scatter collection, with the new colors, sizes and positions.
    #scat.set_edgecolors(rain_drops['color'])
    #scat.set_sizes(rain_drops['size'])
    #scat.set_offsets(rain_drops['position'])


# Construct the animation, using the update function as the animation
# director.
animation = FuncAnimation(fig, update, frames=n_frames/MULT)
#plt.show()
animation.save('flow.gif', dpi=80, writer='imagemagick')



"""DFIG.py
Created by Latha Sethuraman, Katherine Dykes.
Copyright (c) NREL. All rights reserved."""

from openmdao.api import Group, Problem, ExplicitComponent,ExecComp,IndepVarComp,ScipyOptimizeDriver
import numpy as np
from math import pi, cos, sqrt, radians, sin, exp, log10, log, tan, atan


class DFIG(ExplicitComponent):
    """ Estimates overall mass, dimensions and Efficiency of DFIG generator. """
    
    def setup(self):
        # DFIG design inputs
        #self.add_input('r_s', val=0.0, units='m', desc='airgap radius r_s')
        self.add_input('rad_ag', val=0.0, units='m', desc='airgap radius')
        self.add_input('l_s', val=0.0, units='m', desc='Stator core length l_s')
        self.add_input('h_s', val=0.0, units='m', desc='Stator slot height')
        self.add_input('h_r', val=0.0, units='m', desc='Rotor slot height')
        
        self.add_input('machine_rating', val=0.0, units='W',   desc='Machine rating')
        self.add_input('n_nom',          val=0.0, units='rpm', desc='rated speed')
        
        self.add_input('B_symax', val=0.0,            desc='Peak Stator Yoke flux density B_symax')
        self.add_input('S_Nmax',  val=0.0, units='T', desc='Max rated Slip ')
        self.add_input('I_0',     val=0.0, units='A', desc='no-load excitation current')
        
        self.add_input('shaft_cm',           val= np.zeros(3), units='m', desc='Main Shaft CM')
        self.add_input('shaft_length',       val=0.0,          units='m', desc='main shaft length')
        self.add_input('Gearbox_efficiency', val=0.0,                     desc='Gearbox efficiency')
        
        # Material properties
        self.add_input('rho_Fe',     val=0.0, units='kg*m**-3', desc='Magnetic Steel density ')
        self.add_input('rho_Copper', val=0.0, units='kg*m**-3', desc='Copper density ')
        
        # DFIG generator design output
        # Magnetic loading
        self.add_output('B_g',     val=0.0, desc='Peak air gap flux density B_g')
        self.add_output('B_g1',    val=0.0, desc='air gap flux density fundamental ')
        self.add_output('B_rymax', val=0.0, desc='maximum flux density in rotor yoke')
        self.add_output('B_tsmax', val=0.0, desc='maximum tooth flux density in stator')
        self.add_output('B_trmax', val=0.0, desc='maximum tooth flux density in rotor')
        
        #Stator design
        self.add_output('N_s',                val=0.0, desc='Stator turns')
        self.add_output('S',                  val=0.0, desc='Stator slots')
        self.add_output('h_ys',               val=0.0, desc='Stator Yoke height')
        self.add_output('b_s',                val=0.0, desc='stator slot width')
        self.add_output('b_t',                val=0.0, desc='stator tooth width')
        self.add_output('A_Cuscalc',          val=0.0, desc='Stator Conductor cross-section mm^2')
        self.add_output('Slot_aspect_ratio1', val=0.0, desc='Stator slot aspect ratio')
        self.add_output('D_ratio',            val=0.0, desc='Stator diameter ratio')
        self.add_output('q1',                 val=0.0, desc='Slots per pole per phase')
        
        #Rotor design
        self.add_output('N_r',                val=0.0, desc='Rotor turns')
        self.add_output('Q_r',                val=0.0, desc='Rotor slots')
        self.add_output('h_yr',               val=0.0, desc='rotor yoke height')
        self.add_output('b_r',                val=0.0, desc='rotor slot width')
        self.add_output('b_tr',               val=0.0, desc='rotor tooth width')
        self.add_output('A_Curcalc',          val=0.0, desc='Rotor Conductor cross-section mm^2')
        self.add_output('Slot_aspect_ratio2', val=0.0, desc='Rotor slot aspect ratio')
        self.add_output('b_trmin',            val=0.0, desc='minimum tooth width')
        self.add_output('tau_p',              val=0.0, desc='Pole pitch')
        self.add_output('p',                  val=0.0, desc='No of pole pairs')
        
        # Electrical performance
        self.add_output('E_p',  val=0.0, desc='Stator phase voltage')
        self.add_output('f',    val=0.0, desc='Generator output frequency')
        self.add_output('I_s',  val=0.0, desc='Generator output phase current')
        self.add_output('A_1' , val=0.0, desc='Specific current loading')
        self.add_output('J_s',  val=0.0, desc='Stator winding Current density')
        self.add_output('J_r',  val=0.0, desc='Rotor winding Current density')
        self.add_output('R_s',  val=0.0, desc='Stator resistance')
        self.add_output('R_R',  val=0.0, desc='Rotor resistance')
        self.add_output('L_r',  val=0.0, desc='Rotor inductance')
        self.add_output('L_s',  val=0.0, desc='Stator synchronising inductance')
        self.add_output('L_sm', val=0.0, desc='mutual inductance')
        
        # Objective functions
        self.add_output('Mass',    val=0.0, desc='Actual mass')
        self.add_output('K_rad',   val=0.0, desc='Stack length ratio')
        self.add_output('Losses',  val=0.0, desc='Total loss')
        self.add_output('gen_eff', val=0.0, desc='Generator efficiency')
        
        # Mass Outputs
        self.add_output('Copper',          val=0.0, units='kg', desc='Copper Mass')
        self.add_output('Iron',            val=0.0, units='kg', desc='Electrical Steel Mass')
        self.add_output('Structural_mass', val=0.0, units='kg', desc='Structural Mass')
        
        # Structural performance
        self.add_output('TC1', val=0.0, desc='Torque constraint-stator')
        self.add_output('TC2', val=0.0, desc='Torque constraint-rotor')
        
        # Other parameters
        self.add_output('Current_ratio', val=0.0,         desc='Rotor current ratio')
        self.add_output('Overall_eff',   val=0.0,         desc='Overall drivetrain efficiency')
        self.add_output('I',             val=np.zeros(3), desc='Moments of Inertia for the component [Ixx, Iyy, Izz] around its center of mass')
        self.add_output('cm',            val=np.zeros(3), desc='COM [x,y,z]')
        
        #self.declare_partials('*', '*', method='fd', form='central', step=1e-6)

        
    def compute(self, inputs, outputs):
        # Unpack inputs
        rad_ag               = inputs['rad_ag']
        l_s                  = inputs['l_s']
        h_s                  = inputs['h_s']
        h_r                  = inputs['h_r']
        I_0                  = inputs['I_0']
        
        machine_rating       = inputs['machine_rating']
        n_nom                = inputs['n_nom']
        Gearbox_efficiency   = inputs['Gearbox_efficiency']
        
        rho_Fe               = inputs['rho_Fe']
        rho_Copper           = inputs['rho_Copper']
        
        B_symax              = inputs['B_symax']
        S_Nmax               = inputs['S_Nmax']
        
        shaft_cm             = inputs['shaft_cm']
        shaft_length         = inputs['shaft_length']
        
        #Assign values to universal constants
        g1          = 9.81             # m/s^2 acceleration due to gravity
        sigma       = 21.5e3           # shear stress
        mu_0        = pi * 4e-7        # permeability of free space in m * kg / (s**2 * A**2)
        cofi        = 0.9              # power factor
        h_w         = 0.005            # wedge height
        m           = 3                # Number of phases
        resist_Cu   = 1.8e-8 * 1.4     # copper resisitivity
        h_sy0       = 0
        
        #Assign values to design constants
        b_so = 0.004                    # Stator slot opening width
        b_ro = 0.004                    # Rotor  slot opening width
        q1 = 5                          # Stator slots per pole per phase
        q2 = q1 - 1                     # Rotor  slots per pole per phase
        b_s_tau_s = 0.45                # Stator slot-width / slot-pitch ratio
        b_r_tau_r = 0.45                # Rotor  slot-width / slot-pitch ratio
        y_tau_p  = 12. / 15             # Stator coil span to pole pitch    
        y_tau_pr = 10. / 12             # Rotor  coil span to pole pitch
        
        p = 3                           # pole pairs
        freq = 60                       # grid frequency in Hz
        k_fillr = 0.55                  # Rotor Slot fill factor
        k_fills = 0.65                  # Stator Slot fill factor (will be set later)
        P_Fe0h = 4                      # specific hysteresis losses W / kg @ 1.5 T 
        P_Fe0e = 1                      # specific eddy losses W / kg @ 1.5 T 
        
        K_rs     = 1 / ( -1 * S_Nmax)              # Winding turns ratio between rotor and Stator        
        I_SN   = machine_rating / (sqrt(3) * 3000) # Rated current
        I_SN_r = I_SN / K_rs                       # Stator rated current reduced to rotor
        
        # Calculating winding factor for stator and rotor    
        
        k_y1 = sin(pi/2 * y_tau_p)                              # winding chording factor
        k_q1 = sin(pi/6) / (q1 * sin(pi/(6 * q1)))              # winding zone factor
        k_y2 = sin(pi/2 * y_tau_pr)                             # winding chording factor
        k_q2 = sin(pi/6) / (q2 * sin(pi/(6 * q2)))              # winding zone factor
        k_wd1 = k_y1 * k_q1                                     # Stator winding factor
        k_wd2 = k_q2 * k_y2                                     # Rotor winding factor
        
        ag_dia = 2 * rad_ag                                     # air gap diameter
        ag_len = (0.1 + 0.012 * machine_rating**(1./3)) * 0.001 # air gap length in m
        K_rad = l_s / ag_dia                                    # Aspect ratio
        rad_r = rad_ag - ag_len                                 # rotor radius  (was r_r)
        tau_p = pi * ag_dia / (2 * p)                           # pole pitch
        
        S = 2 * p * q1 * m                                      # Stator slots
        N_slots_pp = S / (m * p * 2)                            # Number of stator slots per pole per phase
        n  = S / 2 * p / q1                                     # no of slots per pole per phase
        tau_s = tau_p / (m * q1)                                # Stator slot pitch
        b_s = b_s_tau_s * tau_s                                 # Stator slot width
        b_t = tau_s - b_s                                       # Stator tooth width
        
        Q_r = 2 * p * m * q2                                    # Rotor slots
        tau_r = pi * (ag_dia - 2 * ag_len) / Q_r                # Rotor slot pitch
        b_r = b_r_tau_r * tau_r                                 # Rotor slot width
        b_tr = tau_r - b_r                                      # Rotor tooth width
        
        # Calculating equivalent slot openings
        
        mu_rs = 0.005
        mu_rr = 0.005
        W_s   = (b_s / mu_rs) * 1e-3  # Stator, in m
        W_r   = (b_r / mu_rr) * 1e-3  # Rotor,  in m
        
        Slot_aspect_ratio1 = h_s / b_s
        Slot_aspect_ratio2 = h_r / b_r
        
        # Calculating Carter factor for stator,rotor and effective air gap length
        
        gamma_s = (2 * W_s / ag_len)**2 / (5 + 2 * W_s / ag_len)
        K_Cs    = tau_s / (tau_s - ag_len * gamma_s * 0.5)  #page 3 - 13
        gamma_r = (2 * W_r / ag_len)**2 / (5 + 2 * W_r / ag_len)
        K_Cr    = tau_r / (tau_r - ag_len * gamma_r * 0.5)  #page 3 - 13
        K_C     = K_Cs * K_Cr
        g_eff   = K_C * ag_len
        
        om_m = 2 * pi * n_nom / 60                      # mechanical frequency
        om_e = p * om_m                                 # electrical frequency
        f = n_nom * p / 60                              # generator output freq
        K_s = 0.3                                       # saturation factor for Iron
        n_c1 = 2                                        # number of conductors per coil
        a1 = 2                                          # number of parallel paths
        N_s = np.round(2 * p * N_slots_pp * n_c1 / a1)  # Stator winding turns per phase        
        N_r = np.round(N_s * k_wd1 * K_rs / k_wd2)      # Rotor winding turns per phase
        n_c2 = N_r / (Q_r / m)                          # rotor turns per coil
        
        # Calculating peak flux densities and back iron thickness
        
        B_g1 = mu_0 * 3 * N_r * I_0 * sqrt(2) * k_y2 * k_q2 / (pi * p * g_eff * (1 + K_s))
        B_g = B_g1 * K_C
        h_ys = B_g * tau_p / (B_symax * pi)
        B_rymax = B_symax
        h_yr = h_ys
        B_tsmax = B_g * tau_s / b_t 
        
        d_se = ag_dia + 2 * (h_ys + h_s + h_w)           # stator outer diameter
        D_ratio = d_se / ag_dia                          # Diameter ratio
        
        # Stator slot fill factor
        if ag_dia > 2:
            k_fills = 0.65
        else:
            k_fills = 0.4
            
        # Stator winding calculation
        
        # End connection length for stator winding coils
        
        l_fs = 2 * (0.015 + y_tau_p * tau_p / (2 * cos(radians(40)))) + pi * h_s   # added radians() 2019 09 11
        
        l_Cus = 2 * N_s * (l_fs + l_s) / a1             # Length of Stator winding 
        
        # Conductor cross-section
        A_s       = b_s *        (h_s - h_w)
        A_scalc   = b_s * 1000 * (h_s - h_w) * 1000
        A_Cus     = A_s     * q1 * p * k_fills / N_s
        A_Cuscalc = A_scalc * q1 * p * k_fills / N_s
        
        # Stator winding resistance
        
        R_s = l_Cus * resist_Cu / A_Cus
        tau_r_min = pi * (ag_dia - 2 * (ag_len + h_r)) / Q_r
        
        # Peak magnetic loading on the rotor tooth
        
        b_trmin = tau_r_min - b_r_tau_r * tau_r_min
        B_trmax = B_g * tau_r / b_trmin
        
        # Calculating leakage inductance in  stator
        
        K_01 = 1 - 0.033 * (W_s**2 / ag_len / tau_s)
        sigma_ds = 0.0042
        
        L_ssigmas  = (2 * mu_0 * l_s * n_c1**2 * S / m / a1**2) * ((h_s - h_w) / (3 * b_s) + h_w / b_so)                # slot leakage inductance
        L_ssigmaew = (2 * mu_0 * l_s * n_c1**2 * S / m / a1**2) * 0.34 * q1 * (l_fs - 0.64 * tau_p * y_tau_p) / l_s     # end winding leakage inductance
        L_ssigmag  = (2 * mu_0 * l_s * n_c1**2 * S / m / a1**2) * (0.9 * tau_s * q1 * k_wd1 * K_01 * sigma_ds / g_eff)  # tooth tip leakage inductance
        L_ssigma   = (L_ssigmas + L_ssigmaew + L_ssigmag)                                                               # stator leakage inductance
        L_sm       = 6 * mu_0 * l_s * tau_p * (k_wd1 * N_s)**2 / (pi**2 * (p) * g_eff * (1 + K_s))                      
        L_s        = (L_ssigmas + L_ssigmaew + L_ssigmag)                                                               # stator  inductance
        
        # Calculating leakage inductance in  rotor
        
        K_02 = 1 - 0.033 * (W_r**2 / ag_len / tau_r)
        sigma_dr = 0.0062
        
        l_fr = (0.015 + y_tau_pr * tau_r / 2 / cos(radians(40))) + pi * h_r                                   # Rotor end connection length
        L_rsl = (mu_0 * l_s * (2 * n_c2)**2 * Q_r / m) * ((h_r - h_w) / (3 * b_r) + h_w / b_ro)               # slot leakage inductance
        L_rel = (mu_0 * l_s * (2 * n_c2)**2 * Q_r / m) * 0.34 * q2 * (l_fr - 0.64 * tau_r * y_tau_pr) / l_s   # end winding leakage inductance
        L_rtl = (mu_0 * l_s * (2 * n_c2)**2 * Q_r / m) * (0.9 * tau_s * q2 * k_wd2 * K_02 * sigma_dr / g_eff) # tooth tip leakage inductance
        L_r = (L_rsl + L_rtl + L_rel) / K_rs**2                                                               # rotor leakage inductance
        sigma1 = 1 - (L_sm**2 / L_s / L_r)
        
        #Rotor Field winding
        
        # conductor cross-section
        diff = h_r - h_w
        A_Cur = k_fillr * p*q2 * b_r * diff / N_r
        A_Curcalc = A_Cur * 1e6
        
        L_cur = 2 * N_r * (l_fr + l_s)     # rotor winding length
        Resist_r = resist_Cu * L_cur / A_Cur    # Rotor resistance
        R_R = Resist_r / K_rs**2                # Equivalent rotor resistance reduced to stator
        
        om_s = n_nom * 2*pi / 60                     # synchronous speed in rad / s
        P_e = machine_rating / (1 - S_Nmax)          # Air gap power
        
        # Calculating No-load voltage
        E_p      = om_s * N_s * k_wd1 * rad_ag * l_s * B_g1 * sqrt(2)
        I_r      = P_e / m / E_p                        # rotor active current        
        I_sm     = E_p / (2 * pi * freq * (L_s + L_sm)) # stator reactive current
        I_s      = sqrt(I_r**2 + I_sm**2)               # Stator current
        I_srated = machine_rating / 3 / K_rs / E_p      # Rated current
        
        # Calculating winding current densities and specific current loading
        
        J_s = I_s / A_Cuscalc
        J_r = I_r / A_Curcalc
        A_1 = 2 * m*N_s * I_s / (pi * 2 * rad_ag)
        Current_ratio = I_0 / I_srated           # Ratio of magnetization current to rated current
        
        # Calculating masses of the electromagnetically active materials
        
        V_Cuss = m * l_Cus * A_Cus
        V_Cusr = m * L_cur * A_Cur
        V_Fest = l_s * pi * ((rad_ag + h_s)**2 - rad_ag**2) - (2 * m*q1 * p*b_s * h_s * l_s)
        V_Fesy = l_s * pi * ((rad_ag + h_s + h_ys)**2    - (rad_ag + h_s)**2)
        V_Fert = l_s * pi * (rad_r**2 - (rad_r - h_r)**2) - 2 * m*q2 * p*b_r * h_r * l_s
        V_Fery = l_s * pi * ((rad_r - h_r)**2           - (rad_r - h_r - h_yr)**2)
        Copper = (V_Cuss + V_Cusr) * rho_Copper
        M_Fest = V_Fest * rho_Fe
        M_Fesy = V_Fesy * rho_Fe
        M_Fert = V_Fert * rho_Fe
        M_Fery = V_Fery * rho_Fe
        Iron = M_Fest + M_Fesy + M_Fert + M_Fery
        M_gen = (Copper) + (Iron)
        
        #K_gen = Cu * C_Cu + (Iron) * C_Fe #%M_pm * K_pm
        
        L_tot = l_s
        Structural_mass = 0.0002 * M_gen**2 + 0.6457 * M_gen + 645.24
        Mass = M_gen + Structural_mass
        
        # Calculating Losses and efficiency 
        # 1. Copper losses
        
        K_R = 1.2 # skin effect correction coefficient
        P_Cuss = m * I_s**2 * R_s * K_R            # Copper loss - stator
        P_Cusr = m * I_r**2 * R_R                  # Copper loss - rotor
        P_Cusnom = P_Cuss + P_Cusr                 # Copper loss - total
        
        # Iron Losses ( from Hysteresis and eddy currents)      
        P_Hyys = M_Fesy * (B_symax / 1.5)**2 * (P_Fe0h * om_e / (2 * pi * 60))                    # Hysteresis losses in stator yoke
        P_Ftys = M_Fesy * (B_symax / 1.5)**2 * (P_Fe0e * (om_e / (2 * pi * 60))**2)               # Eddy losses in stator yoke
        P_Hyd  = M_Fest * (B_tsmax / 1.5)**2 * (P_Fe0h * om_e / (2 * pi * 60))                    # Hysteresis losses in stator teeth
        P_Ftd  = M_Fest * (B_tsmax / 1.5)**2 * (P_Fe0e * (om_e / (2 * pi * 60))**2)               # Eddy losses in stator teeth
        P_Hyyr = M_Fery * (B_rymax / 1.5)**2 * (P_Fe0h * abs(S_Nmax) * om_e / (2 * pi * 60))      # Hysteresis losses in rotor yoke
        P_Ftyr = M_Fery * (B_rymax / 1.5)**2 * (P_Fe0e * (abs(S_Nmax) * om_e / (2 * pi * 60))**2) # Eddy losses in rotor yoke
        P_Hydr = M_Fert * (B_trmax / 1.5)**2 * (P_Fe0h * abs(S_Nmax) * om_e / (2 * pi * 60))      # Hysteresis losses in rotor teeth
        P_Ftdr = M_Fert * (B_trmax / 1.5)**2 * (P_Fe0e * (abs(S_Nmax) * om_e / (2 * pi * 60))**2) # Eddy losses in rotor teeth
        P_add = 0.5 * machine_rating / 100                                                        # additional losses
        P_Fesnom = P_Hyys + P_Ftys + P_Hyd + P_Ftd + P_Hyyr + P_Ftyr + P_Hydr + P_Ftdr            # Total iron loss
        delta_v = 1                                                                               # allowable brush voltage drop
        p_b = 3 * delta_v * I_r                                                                   # Brush loss
        
        Losses = P_Cusnom + P_Fesnom + p_b + P_add
        gen_eff = (P_e - Losses) * 100 / P_e
        Overall_eff = gen_eff * Gearbox_efficiency
        
        # Calculating stator winding current density
        J_s = I_s / A_Cuscalc
        
        # Calculating  electromagnetic torque
        T_e = p *(machine_rating * 1.01) / (2 * pi * freq * (1 - S_Nmax))
        
        # Calculating for tangential stress constraints
        
        TC1 = T_e / (2 * pi * sigma)
        TC2 = rad_ag**2 * l_s
        
        # Calculating mass moments of inertia and center of mass
        I = np.zeros(3)
        r_out = d_se * 0.5
        I[0]   = (0.5 * Mass * r_out**2)
        I[1]   = (0.25 * Mass * r_out**2 + Mass * l_s**2 / 12) 
        I[2]   = I[1]
        cm = np.zeros(3)
        cm[0]  = shaft_cm[0] + shaft_length/2. + l_s/2.
        cm[1]  = shaft_cm[1]
        cm[2]  = shaft_cm[2]

        outputs['B_g']                = B_g
        outputs['B_g1']               = B_g1
        outputs['B_rymax']            = B_rymax
        outputs['B_tsmax']            = B_tsmax
        outputs['B_trmax']            = B_trmax

        outputs['q1']                 = q1
        outputs['N_s']                = N_s
        outputs['S']                  = S
        outputs['h_ys']               = h_ys
        outputs['b_s']                = b_s
        outputs['b_t']                = b_t

        outputs['D_ratio']            = D_ratio
        outputs['A_Cuscalc']          = A_Cuscalc
        outputs['Slot_aspect_ratio1'] = Slot_aspect_ratio1
        outputs['h_yr']               = h_yr

        outputs['tau_p']              = tau_p
        outputs['p']                  = p
        outputs['Q_r']                = Q_r
        outputs['N_r']                = N_r
        outputs['b_r']                = b_r
        outputs['b_trmin']            = b_trmin

        outputs['b_tr']               = b_tr
        outputs['A_Curcalc']          = A_Curcalc
        outputs['Slot_aspect_ratio2'] = Slot_aspect_ratio2
        outputs['E_p']                = E_p
        outputs['f']                  = f
        
        outputs['I_s']                = I_s
        outputs['A_1']                = A_1
        outputs['J_s']                = J_s
        outputs['J_r']                = J_r
        outputs['R_s']                = R_s
        outputs['R_R']                = R_R

        outputs['L_r']                = L_r
        outputs['L_s']                = L_s
        outputs['L_sm']               = L_sm
        outputs['Mass']               = Mass
        outputs['K_rad']              = K_rad
        outputs['Losses']             = Losses

        outputs['gen_eff']            = gen_eff
        outputs['Copper']             = Copper
        outputs['Iron']               = Iron
        outputs['Structural_mass']    = Structural_mass
        outputs['TC1']                = TC1
        outputs['TC2']                = TC2

        outputs['Current_ratio']      = Current_ratio
        outputs['Overall_eff']        = Overall_eff
        outputs['cm']                 = cm
        outputs['I']                  =  I
        
        



import unittest

from wisdem.test.test_floatingse import test_column
from wisdem.test.test_floatingse import test_map_mooring
from wisdem.test.test_floatingse import test_loading
from wisdem.test.test_floatingse import test_substructure
from wisdem.test.test_floatingse import test_floating

def suite():
    suite = unittest.TestSuite( (test_column.suite(),
                                 test_map_mooring.suite(),
                                 test_loading.suite(),
                                 test_substructure.suite(),
                                 test_floating.suite()
    ) )
    return suite


if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())
        
